# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np

try:
    import pandaplan.core.pplog as pplog
except ImportError:
    import logging as pplog

from pandapower.auxiliary import NetCalculationNotConverged, ControllerNotConverged
from pandapower.powerflow import LoadflowNotConverged
from pandapower.optimal_powerflow import OPFNotConverged
from pandapower.control.util.auxiliary import asarray

logger = pplog.getLogger(__name__)

from pandapower.run import runpp


def get_controller_order(nets, controller):
    """
    Defining the controller order per level
    Takes the order and level columns from net.controller
    If levels are specified, the levels and orders are executed in ascending order.
    :param net:
    :return: level_list - list of levels to be run
    :return: controller_order - list of controller order lists per level
    """
    # let level be float so that a new level can be added in between of existing ones
    if not hasattr(nets, '__len__') or (isinstance(nets, dict)):
        nets = [nets] * len(controller)
        nets = np.array(nets)
    if nets is not np.ndarray:
        nets = np.array(nets)
    if np.any(controller.level.isnull()):
        raise UserWarning("controller level cannot be None")
    level = controller.level.apply(asarray).values
    # list of sorted unique levels
    level_list = sorted(set(np.concatenate(level)))

    # identify the levels to be run
    controller_order = []
    for l in level_list:
        to_add = controller.in_service.values & [*map(lambda x: l in x, level)]
        rel_controller, order = controller['object'].values[to_add], controller['order'].values[to_add]
        controller_order.append([*zip(rel_controller[order.argsort()], nets[to_add][order.argsort()])])
        # controller_order.append(net.controller[to_add].sort_values(["order"]).object.values)

    if logger.level <= pplog.DEBUG:
        logger.debug("levellist: " + str(level_list))
        logger.debug("order: " + str(controller_order)) # Note: creates a long string if many controllers are present

    return level_list, controller_order


def check_for_initial_run(controller_order):
    """
    Function checking if any of the controllers need an initial power flow
    If net has no controllers, an initial power flow is done by default
    """

    if not len(controller_order[0]):
        return True

    for levelorder in controller_order:
        for ctrl, net in levelorder:
            if hasattr(ctrl, 'initial_powerflow'):
                net.controller.at[ctrl.index, 'initial_run']= ctrl.initial_powerflow
                logger.warning("initial_powerflow is deprecated. Instead of defining initial_powerflow "
                               "please define initial_run in the future.")
                del ctrl.initial_powerflow
            elif hasattr(ctrl, 'initial_run'):
                net.controller.at[ctrl.index, 'initial_run']= ctrl.initial_run
                logger.warning("initial_run as attribute is deprecated. initial_run is now part of the "
                               "net.controller DataFrame")
                del ctrl.initial_run
            if net.controller.at[ctrl.index, 'initial_run']:
                return True
    return False


def ctrl_variables_default(net, **kwargs):
    ctrl_variables = dict()
    if not hasattr(net, "controller") or len(net.controller[net.controller.in_service]) == 0:
        ctrl_variables["level"], ctrl_variables["controller_order"] = [0], [[]]
    else:
        ctrl_variables["level"], ctrl_variables["controller_order"] = get_controller_order(net, net.controller)
    ctrl_variables["run"] = kwargs.pop('run', runpp)
    ctrl_variables["initial_run"] = check_for_initial_run(ctrl_variables["controller_order"])
    ctrl_variables['continue_on_divergence'] = False
    ctrl_variables['check_each_level'] = True
    ctrl_variables["errors"] = (LoadflowNotConverged, OPFNotConverged, NetCalculationNotConverged)
    return ctrl_variables


def prepare_run_ctrl(net, ctrl_variables, **kwargs):
    """
    Prepares run control functions. Internal variables needed:

    **controller_order** (list) - Order in which controllers in net.controller will be called
    **runpp** (function) - the runpp function (for time series a faster version is possible)
    **initial_run** (bool) - some controllers need an initial run of the powerflow prior to the control step

    """
    # sort controller_order by order if not already done


    ctrl_var = ctrl_variables

    if ctrl_variables is None:
        ctrl_variables = ctrl_variables_default(net, **kwargs)

    if ('continue_on_divergence') in kwargs and (ctrl_var is None or 'continue_on_divergence' not in ctrl_var.keys()):
        div = kwargs.pop('continue_on_divergence')
        ctrl_variables['continue_on_divergence'] = div
    if ('check_each_level') in kwargs and (ctrl_var is None or 'check_each_level' not in ctrl_var.keys()):
        check = kwargs.pop('check_each_level')
        ctrl_variables['check_each_level'] = check

    return ctrl_variables


def check_final_convergence(run_count, max_iter, net_converged):
    if run_count > max_iter:
        raise ControllerNotConverged("Maximum number of iterations per controller is reached. "
                                     "Some controller did not converge after %i calculations!"
                                     % run_count)
    if not net_converged:
        raise NetCalculationNotConverged("Controller did not converge because the calculation did not converge!")
    else:
        logger.debug("Converged after %i calculations" % run_count)


def get_recycle(ctrl_variables):
    # check if recycle is in ctrl_variables
    recycle, only_v_results = None, False
    if ctrl_variables is not None and "recycle_options" in ctrl_variables:
        recycle = ctrl_variables.get("recycle_options", None)
        if isinstance(recycle, dict):
            only_v_results = recycle.get("only_v_results", False)
    return recycle, only_v_results


def net_initialization(net, ctrl_variables, **kwargs):
    # initial power flow (takes time, but is not needed for every kind of controller)
    initial_run = ctrl_variables['initial_run']
    run_funct = ctrl_variables['run']
    if initial_run:
        run_funct(net, **kwargs)  # run can be runpp, runopf or whatever
    else:
        net["converged"] = True  # assume that the initial state is valid
    ctrl_variables['converged'] = net['converged'] or net['OPF_converged']
    return ctrl_variables


def control_initialization(controller_order):
    # initialize each controller prior to the first power flow
    for levelorder in controller_order:
        for ctrl, net in levelorder:
            ctrl.initialize_control(net)


def _evaluate_net(net, levelorder, ctrl_variables, **kwargs):
    run_funct = ctrl_variables['run']
    errors = ctrl_variables['errors']
    try:
        run_funct(net, **kwargs)  # run can be runpp, runopf or whatever
    except errors as err:
        net._ppc = None
        if ctrl_variables['continue_on_divergence']:
            # give a chance to controllers to "repair" the control step if load flow
            # didn't converge
            # either implement this in a controller that is likely to cause the error,
            # or define a special "load flow police" controller for your use case
            _control_repair(levelorder)
            # this will raise the error if repair_control did't work
            # it means that repair control has only 1 try
            try:
                run_funct(net, **kwargs)
            except errors:
                pass
        else:
            raise err
    ctrl_variables['converged'] = net['converged'] or net.get('OPF_converged', False)
    return ctrl_variables


def control_implementation(net, controller_order, ctrl_variables, max_iter,
                           evaluate_net_fct=_evaluate_net, **kwargs):

    run_count=0
    # run each controller step in given controller order
    for levelorder in controller_order:
        _reset_convergence(levelorder)
        # converged gives status about convergence of a controller. Is initialized as False
        ctrl_converged = False
        # run_count is 0 before entering the loop. Is incremented in each controller loop
        converged = ctrl_variables['converged']
        run_count = 0
        while not ctrl_converged and run_count <= max_iter and converged:
            ctrl_converged = _control_step(levelorder, run_count)
            # call to run function (usually runpp) after each controller was called
            # this function is called at least once per level
            if not ctrl_converged:
                run_count += 1
                ctrl_variables = evaluate_net_fct(net, levelorder, ctrl_variables, **kwargs)
        # raises controller not converged
        if ctrl_variables['check_each_level']:
            check_final_convergence(run_count, max_iter, ctrl_variables['converged'])
    # is required if you only want to check if in the last level everything is converged
    check_final_convergence(run_count, max_iter, ctrl_variables['converged'])

def _reset_convergence(levelorder):
    for ctrl, net in levelorder:
        ctrl.level_reset(net)


def _control_step(levelorder, run_count):
    # keep track of stopping criteria
    converged = True
    logger.debug("Controller Iteration #%i" % run_count)
    # run each controller until all are converged
    for ctrl, net in levelorder:
        # call control step while controller ist not converged yet
        if not ctrl.is_converged(net):
            ctrl.control_step(net)
            converged = False
    return converged


def _control_repair(levelorder):
    for ctrl, net in levelorder:
        ctrl.repair_control(net)


def control_finalization(controller_order):
    # call finalize function of each controller
    for levelorder in controller_order:
        for ctrl, net in levelorder:
            ctrl.finalize_control(net)


def run_control(net, ctrl_variables=None, max_iter=30, **kwargs):
    """
    Main function to call a net with controllers
    Function is running control loops for the controllers specified in net.controller

    INPUT:
        **net** - pandapower network with controllers included in net.controller

    OPTIONAL:
        **ctrl_variables** (dict, None) - variables needed internally to calculate the power flow. See prepare_run_ctrl()

        **max_iter** (int, 30) - The maximum number of iterations for controller to converge

    KWARGS:
        **continue_on_divergence** (bool, False) - if run_funct is not converging control_repair is fired \
                                                   (only relevant if ctrl_varibales is None, otherwise it needs \
                                                   to be defined in ctrl_variables anyway)

        **check_each_level** (bool, True) - if each level shall be checked if the controllers are converged or not \
                                           (only relevant if ctrl_varibales is None, otherwise it needs \
                                           to be defined in ctrl_variables anyway)

    Runs controller until each one converged or max_iter is hit.

    1. Call initialize_control() on each controller
    2. Calculate an inital power flow (if it is enabled, i.e. setting the initial_run veriable to True)
    3. Repeats the following steps in ascending order of controller_order until total convergence of all
       controllers for each level:

        a) Evaluate individual convergence for all controllers in the level
        b) Call control_step() for all controllers in the level on diverged controllers
        c) Calculate power flow (or optionally another function like runopf or whatever you defined)
    4. Call finalize_control() on each controller

    """
    ctrl_variables = prepare_run_ctrl(net, ctrl_variables)
    kwargs["recycle"], kwargs["only_v_results"] = get_recycle(ctrl_variables)

    controller_order = ctrl_variables["controller_order"]

    # initialize each controller prior to the first power flow
    control_initialization(controller_order)

    # initial power flow (takes time, but is not needed for every kind of controller)
    ctrl_variables = net_initialization(net, ctrl_variables, **kwargs)

    # run each controller step in given controller order
    control_implementation(net, controller_order, ctrl_variables, max_iter, **kwargs)

    # call finalize function of each controller
    control_finalization(controller_order)



import logging
import time

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.staticVarCompensatorCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class StaticVarCompensatorCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_static_var_compensator_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting StaticVarCompensator.")
        eq_stat_coms = self._prepare_static_var_compensator_cim16()
        self.cimConverter.copy_to_pp('shunt', eq_stat_coms)
        self.logger.info("Created %s generators in %ss." % (eq_stat_coms.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s generators from StaticVarCompensator in %ss." %
                    (eq_stat_coms.index.size, time.time() - time_start)))

    def _prepare_static_var_compensator_cim16(self) -> pd.DataFrame:
        eq_stat_coms = self.cimConverter.merge_eq_ssh_profile('StaticVarCompensator', True)
        eq_stat_coms = pd.merge(eq_stat_coms, self.cimConverter.bus_merge, how='left', on='rdfId')
        eq_stat_coms = eq_stat_coms.rename(columns={'q': 'q_mvar'})
        # get the active power and reactive power from SV profile
        eq_stat_coms = pd.merge(eq_stat_coms, self.cimConverter.cim['sv']['SvPowerFlow'][['p', 'q', 'Terminal']],
                                how='left', left_on='rdfId_Terminal', right_on='Terminal')
        eq_stat_coms['q_mvar'] = eq_stat_coms['q_mvar'].fillna(eq_stat_coms['q'])
        eq_stat_coms.loc[eq_stat_coms['sVCControlMode'] == 'reactivePower', 'voltageSetPoint'] = float('NaN')
        if 'inService' in eq_stat_coms.columns:
            eq_stat_coms['connected'] = eq_stat_coms['connected'] & eq_stat_coms['inService']
        eq_stat_coms = eq_stat_coms.rename(columns={'rdfId_Terminal': sc['t'], 'rdfId': sc['o_id'], 'p': 'p_mw',
                                                    'voltageSetPoint': 'vn_kv', 'index_bus': 'bus', 'connected': 'in_service'})
        eq_stat_coms['step'] = 1
        eq_stat_coms['max_step'] = 1
        # create step_dependency_table flag
        if 'step_dependency_table' not in eq_stat_coms.columns:
            eq_stat_coms["step_dependency_table"] = False
        return eq_stat_coms



import logging
import time
import math

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.energySourceCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class EnergySourceCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_energy_sources_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting EnergySources.")
        eqssh_energy_sources = self._prepare_energy_sources_cim16()
        es_slack = eqssh_energy_sources.loc[eqssh_energy_sources.vm_pu.notna()]
        es_sgen = eqssh_energy_sources.loc[eqssh_energy_sources.vm_pu.isna()]
        # create reactive_capability_curve flag
        if 'reactive_capability_curve' not in es_sgen.columns:
            es_sgen['reactive_capability_curve'] = False
        if 'controllable' not in es_sgen.columns:
            es_sgen['controllable'] = False
        self.cimConverter.copy_to_pp('ext_grid', es_slack)
        self.cimConverter.copy_to_pp('sgen', es_sgen)
        # self._copy_to_pp('sgen', eqssh_energy_sources)
        self.logger.info("Created %s sgens in %ss." % (eqssh_energy_sources.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s sgens from EnergySources in %ss." %
                    (eqssh_energy_sources.index.size, time.time() - time_start)))

    def _prepare_energy_sources_cim16(self) -> pd.DataFrame:
        eq_energy_scheduling_type = \
            pd.concat([self.cimConverter.cim['eq']['EnergySchedulingType'],
                       self.cimConverter.cim['eq_bd']['EnergySchedulingType']],
                      sort=False)
        eq_energy_scheduling_type = eq_energy_scheduling_type.rename(
            columns={'rdfId': 'EnergySchedulingType', 'name': 'type'})
        eqssh_energy_sources = self.cimConverter.merge_eq_ssh_profile('EnergySource', add_cim_type_column=True)
        eqssh_energy_sources = pd.merge(eqssh_energy_sources, eq_energy_scheduling_type, how='left',
                                        on='EnergySchedulingType')
        eqssh_energy_sources = pd.merge(eqssh_energy_sources, self.cimConverter.bus_merge, how='left', on='rdfId')
        eqssh_energy_sources = eqssh_energy_sources.drop_duplicates(['rdfId'], keep='first')
        sgen_type = dict({'WP': 'WP', 'Wind': 'WP', 'PV': 'PV', 'SolarPV': 'PV', 'BioGas': 'BioGas',
                          'OtherRES': 'OtherRES', 'CHP': 'CHP'})  # todo move?
        eqssh_energy_sources['type'] = eqssh_energy_sources['type'].map(sgen_type)
        eqssh_energy_sources['p_mw'] = -eqssh_energy_sources['activePower']
        eqssh_energy_sources['q_mvar'] = -eqssh_energy_sources['reactivePower']
        eqssh_energy_sources['va_degree'] = eqssh_energy_sources['voltageAngle'] * 180 / math.pi
        eqssh_energy_sources['vm_pu'] = \
            eqssh_energy_sources['voltageMagnitude'] / eqssh_energy_sources['base_voltage_bus']
        eqssh_energy_sources['scaling'] = 1.
        eqssh_energy_sources['current_source'] = True
        eqssh_energy_sources['generator_type'] = 'current_source'
        if 'inService' in eqssh_energy_sources.columns:
            eqssh_energy_sources['connected'] = (eqssh_energy_sources['connected']
                                                 & eqssh_energy_sources['inService'])
        eqssh_energy_sources = eqssh_energy_sources.rename(columns={'rdfId_Terminal': sc['t'], 'rdfId': sc['o_id'],
                                                                    'connected': 'in_service', 'index_bus': 'bus'})
        return eqssh_energy_sources



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np

from pandapower.auxiliary import read_from_net, _detect_read_write_flag
from pandapower.control.basic_controller import Controller

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)


class TrafoController(Controller):
    """
    Trafo Controller with local tap changer voltage control.

    INPUT:
       **net** (attrdict) - Pandapower struct

       **element_index** (int) - ID of the trafo that is controlled

       **side** (string) - Side of the transformer where the voltage is controlled ("hv", "mv"
       or "lv")

       **tol** (float) - Voltage tolerance band at bus in Percent

       **in_service** (bool) - Indicates if the element is currently active

       **element** (string) - Type of the controlled trafo ("trafo" or "trafo3w")

    OPTIONAL:
        **recycle** (bool, True) - Re-use of internal-data in a time series loop.
    """

    def __init__(self, net, element_index, side, tol, in_service, element, level=0, order=0,
                 recycle=True, **kwargs):
        super().__init__(net, in_service=in_service, level=level, order=order, recycle=recycle,
                         **kwargs)

        self.element = element
        self.element_index = element_index

        self._set_side(side)
        self._set_read_write_flag(net)
        # self._set_valid_controlled_index_and_bus(net)
        self._set_tap_parameters(net)
        self._set_tap_side_coeff(net)

        self.tol = tol

        self.set_recycle(net)

        self.trafobus = None

    def _set_read_write_flag(self, net):
        # if someone changes indices of the controller from single index to array and vice versa
        self._read_write_flag, _ = _detect_read_write_flag(net, self.element, self.element_index, "tap_pos")
        if self._read_write_flag == 'loc':
            self.element_index = np.array(self.element_index)

    def initialize_control(self, net):
        # in case changes applied to net in the meantime:
        # the update occurs in case the in_service parameter of tranformers is changed in the meantime
        # update valid trafo and bus
        # update trafo tap parameters
        # we assume side does not change after the controller is created
        self._set_read_write_flag(net)
        self._update_trafobus(net)
        # self._set_valid_controlled_index_and_bus(net)
        if self.nothing_to_do(net):
            return
        self._set_tap_parameters(net)
        self._set_tap_side_coeff(net)

    def _update_trafobus(self, net):
        self.trafobus = read_from_net(net, self.element, self.element_index, self.side + '_bus', self._read_write_flag)

    def nothing_to_do(self, net):
        element_in_service = read_from_net(net, self.element, self.element_index, 'in_service', self._read_write_flag)
        ext_grid_bus = np.isin(self.trafobus, net.ext_grid.loc[net.ext_grid.in_service, 'bus'].values)
        element_index_in_net = np.isin(self.element_index, net[self.element].index.values)
        self.controlled = np.logical_and(np.logical_and(element_in_service, element_index_in_net),
                                         np.logical_not(ext_grid_bus))
        if isinstance(self.element_index, np.int64) or isinstance(self.element_index, int):
            # if the controller shouldn't do anything, return True
            if not element_in_service or ext_grid_bus or not element_index_in_net or (
                    self._read_write_flag != 'single_index' and len(self.element_index) == 0):
                return True
            return False
        else:
            # if the controller shouldn't do anything, return True
            if np.all(~element_in_service[self.controlled]) or np.all(ext_grid_bus[self.controlled]) or np.all(
                    ~element_index_in_net[self.controlled]) or (
                    self._read_write_flag != 'single_index' and len(self.element_index) == 0):
                return True
            return False

    def _set_tap_side_coeff(self, net):
        tap_side = read_from_net(net, self.element, self.element_index, 'tap_side', self._read_write_flag)
        if (len(np.setdiff1d(tap_side, ['hv', 'lv'])) > 0 and self.element == "trafo") or \
                (len(np.setdiff1d(tap_side, ['hv', 'lv', 'mv'])) > 0 and self.element == "trafo3w"):
            raise ValueError("Trafo tap side (in net.%s) has to be either hv or lv, "
                             "but received: %s for trafo %s" % (self.element, tap_side, self.element_index))

        if self._read_write_flag == "single_index":
            self.tap_side_coeff = 1 if tap_side == 'hv' else -1
            if self.side == "hv":
                self.tap_side_coeff *= -1
            if self.tap_step_percent < 0:
                self.tap_side_coeff *= -1
        else:
            self.tap_side_coeff = np.where(tap_side == 'hv', 1, -1)
            self.tap_side_coeff[self.side == "hv"] *= -1
            self.tap_side_coeff[self.tap_step_percent < 0] *= -1

    def _set_side(self, side):
        if self.element == "trafo":
            if side not in ["hv", "lv"]:
                raise UserWarning("side has to be 'hv' or 'lv' for high/low voltage, "
                                  "received %s" % side)
        elif self.element == "trafo3w":
            if side not in ["hv", "mv", "lv"]:
                raise UserWarning("side has to be 'hv', 'mv' or 'lv' for high/middle/low voltage, "
                                  "received %s" % side)
        else:
            raise UserWarning("unknown trafo type, received %s" % self.element)

        self.side = side

    def _set_valid_controlled_index_and_bus(self, net):
        element_in_service = read_from_net(net, self.element, self.element_index, 'in_service', self._read_write_flag)
        ext_grid_bus = np.isin(self.trafobus, net.ext_grid.loc[net.ext_grid.in_service, 'bus'].values)
        element_index_in_net = np.isin(self.element_index, net[self.element].index.values)
        self.controlled = np.logical_and(np.logical_and(element_in_service, element_index_in_net),
                                         np.logical_not(ext_grid_bus))
        if self._read_write_flag != 'single_index':
            self.element_index = self.element_index[self.controlled]
            self.trafobus = self.trafobus[self.controlled]

        if np.all(~self.controlled):
            logger.warning("All controlled buses are not valid: controller has no effect")

    def _set_tap_parameters(self, net):
        self.tap_min = read_from_net(net, self.element, self.element_index, "tap_min", self._read_write_flag)
        self.tap_max = read_from_net(net, self.element, self.element_index, "tap_max", self._read_write_flag)
        self.tap_neutral = read_from_net(net, self.element, self.element_index, "tap_neutral", self._read_write_flag)
        self.tap_step_percent = read_from_net(net, self.element, self.element_index, "tap_step_percent",
                                              self._read_write_flag)
        self.tap_step_degree = read_from_net(net, self.element, self.element_index, "tap_step_degree",
                                             self._read_write_flag)

        self.tap_pos = read_from_net(net, self.element, self.element_index, "tap_pos", self._read_write_flag)
        if self._read_write_flag == "single_index":
            self.tap_sign = 1 if np.isnan(self.tap_step_degree) else np.sign(np.cos(np.deg2rad(self.tap_step_degree)))
            if (self.tap_sign == 0) | (np.isnan(self.tap_sign)):
                self.tap_sign = 1
            if np.isnan(self.tap_pos):
                self.tap_pos = self.tap_neutral
        else:
            self.tap_sign = np.where(np.isnan(self.tap_step_degree), 1,
                                     np.sign(np.cos(np.deg2rad(self.tap_step_degree))))
            self.tap_sign = np.where((self.tap_sign == 0) | (np.isnan(self.tap_sign)), 1, self.tap_sign)
            self.tap_pos = np.where(np.isnan(self.tap_pos), self.tap_neutral, self.tap_pos)

        if np.any(np.isnan(self.tap_min)) or np.any(np.isnan(self.tap_max)) or np.any(np.isnan(self.tap_step_percent)):
            logger.error("Trafo-Controller has been initialized with NaN values, check "
                         "net.trafo.tap_pos etc. if they are set correctly!")

    def set_recycle(self, net):
        allowed_elements = ["trafo", "trafo3w"]
        if net.controller.at[self.index, 'recycle'] is False or self.element not in allowed_elements:
            # if recycle is set to False by the user when creating the controller it is deactivated or when
            # const control controls an element which is not able to be recycled
            net.controller.at[self.index, 'recycle'] = False
            return
        # these variables determine what is re-calculated during a time series run
        recycle = dict(trafo=True, gen=False, bus_pq=False)
        net.controller.at[self.index, 'recycle'] = recycle

    # def timestep(self, net):
    #     self.tap_pos = net[self.element].at[self.element_index, "tap_pos"]

    def __repr__(self):
        s = '%s of %s %s' % (self.__class__.__name__, self.element, self.element_index)
        return s

    def __str__(self):
        s = '%s of %s %s' % (self.__class__.__name__, self.element, self.element_index)
        return s



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: mrichter
"""
from __future__ import annotations
from typing import List, Dict, Union
import pandas as pd
import json


class ReportContainer:
    pass


class CustomJSONEncoder(json.JSONEncoder):
    pass


class CustomLogger(CustomJSONEncoder):
    pass


class AssetType(CustomLogger):
    pass


class RepairMain(CustomLogger):
    def __init__(self, assets: List[str] = None):
        super().__init__()
        self.assets = assets

    def get_assets(self):
        pass

    def repair(self, data: Dict[str, pd.DataFrame], report_container: ReportContainer = None):
        pass

    def deserialize(self, path_or_json_str: str, report_container: ReportContainer) -> RepairMain:
        pass


class PandapowerRepair(RepairMain):
    def __init__(self, assets: List[AssetType] = None):
        super().__init__(assets)

    def deserialize(self, path_or_json_str: Union[str, PandapowerRepair], report_container: ReportContainer = None) -> \
            PandapowerRepair:
        pass


class CIMRepair(CustomLogger):
    def __init__(self, profiles: Dict[str, RepairMain] = None):
        super().__init__()
        self.profiles = profiles

    def set_profile(self, profile: str, repair_main: RepairMain):
        pass

    def repair(self, cim_dict: Dict[str, Dict[str, pd.DataFrame]], report_container: ReportContainer = None):
        pass

    def deserialize(self, path_or_json_str: Union[str, CIMRepair], report_container: ReportContainer = None) -> \
            CIMRepair:
        pass



import time
from copy import deepcopy

from pandapower.create import create_group_from_dict
from pandapower.grid_equivalents.auxiliary import drop_assist_elms_by_creating_ext_net, \
    drop_internal_branch_elements, add_ext_grids_to_boundaries, \
    _ensure_unique_boundary_bus_names, match_cost_functions_and_eq_net, \
    _check_network, _runpp_except_voltage_angles
from pandapower.grid_equivalents.rei_generation import _create_net_zpbn, \
    _get_internal_and_external_nets, _calculate_equivalent_Ybus, \
    _create_bus_lookups, _calclate_equivalent_element_params, \
    _replace_ext_area_by_impedances_and_shunts
from pandapower.grid_equivalents.ward_generation import \
    _calculate_ward_and_impedance_parameters, \
    _calculate_xward_and_impedance_parameters, \
    create_passive_external_net_for_ward_admittance, \
    _replace_external_area_by_wards, _replace_external_area_by_xwards
from pandapower.groups import isin_group, set_group_reference_column
from pandapower.run import runpp
from pandapower.toolbox.data_modification import reindex_buses
from pandapower.toolbox.grid_modification import replace_ward_by_internal_elements, replace_xward_by_internal_elements, \
    drop_buses, drop_elements_at_buses, merge_nets, fuse_buses
from pandapower.topology.create_graph import create_nxgraph
from pandapower.topology.graph_searches import connected_component, connected_components

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def get_equivalent(net, eq_type, boundary_buses, internal_buses,
                   return_internal=True, show_computing_time=False,
                   ward_type="ward_injection", adapt_va_degree=False,
                   calculate_voltage_angles=True,
                   allow_net_change_for_convergence=False,
                   runpp_fct=_runpp_except_voltage_angles, **kwargs):
    """
    This function calculates and implements the rei or ward/xward network
    equivalents.

    ATTENTION:

        - Power flow results in the given pandapower net are mandatory.

    Known REI equivalents problems:

        - shift_degree != 0 of trafos and trafo3ws lead to errors or wrong results

        - despite 'adapt_va_degree', va_degree at the slack bus showed deviations within large grids

        - with large, real grids small deviations in the power flow results occured, in small grids \
            the results fit perfectly

    INPUT:
        **net** - The pandapower network including power flow results will not be changed during this function.

        **eq_type** (string) - type of the eqvalent network

            The following methods are available:

                - "rei": The idea of rei equivalent is to aggregate the power \
                        and current injection of the external buses to one or \
                        more fictitious radial, equivalent and independent \
                        (rei) nodes. There are three types of the rei-node in \
                        this routine, i.e. the reserved load, the reserved sgen \
                        and the reserved gen (also ext_grid). According to the \
                        demand, these elements (load, sgen and gen) are \
                        considered in the function "_create_net_zpbn" \
                        integrated or separately.

                - "ward": The ward-type equivalent represents the external \
                        network with some equivalent impedance, shunts and \
                        power injections at boundary buses. The equivalent \
                        power injections represent the power of the elements \
                        (load, sgen and gen), which are removed by the \
                        equivalent calculation.

                - "xward": The xward equivalent is an extended variation of \
                        the ward equivalent. Addition to the \
                        ward-representation, a fictitious PV node (generator) \
                        is added with zero active power injection at every \
                        boundary bus. The voltage of the PV node is set \
                        according to the boundary bus voltage.

                ward/xward has two mothods to develop an equivalent circuits, \
                i.e. the injection method and the admittance method. In the \
                admittance method, all the bus power injections in the external \
                networks are converted into shunt admittance before network \
                equivalent. That is the only difference between these two \
                methods. With the help of the function "adapt_net", these \
                methods are configurable.

        **boundary_buses** (iterable) - list of boundary bus indices, by which \
            the original network are divided into two networks, i.e. an \
            internal network and an external network.

        **internal_buses** (iterable) - list of bus indices, which are within \
            the internal network. The program will add buses which are \
            connected to this list of internal buses without passing boundary \
            buses. If 'internal_buses' is an empty list or None, the whole \
            grid is treated as external network.

    OPTIONAL:
        **return_internal** (bool, True) - Reservation of the internal network

             If True, the internal network is reserved in the final equivalent \
             network; otherwise only the external network is the output.

        **show_computing_time** (bool, False) - show computing time of each step

        **ward_type** (str, "ward_injection") - Type of ward and xward

            ward and xward proposed two mothods called the Ward Injection \
            method and the Ward Admittance method to develop equivalent \
            circuits. The only difference between these methods is that in \
            the Ward Admittance method, all bus power in the external networks \
            are converted into shunt admittances before network equivalent.

        **adapt_va_degree** (bool, None) - if True, in add_ext_grids_to_boundaries(), the va_degree \
            of the additional ext_grids (if needed) at the boundary buses will be increased or \
            decreased to values that minimize the difference to the given res_bus.va_degree values.

        **allow_net_change_for_convergence** (bool, False) - if the net doesn't converge at the \
            first internal power flow, which is in add_ext_grids_to_boundaries(), and this parameter is \
            True, the code tests if changes to unusual impedance values solve the divergence issue.

        **calculate_voltage_angles** (bool, True) - parameter passed to internal runpp() runs.

        ****kwargs** - key word arguments, such as sgen_separate, load_separate, gen_separate, \
        group_name.

    OUTPUT:
         **net_eq** - The equivalent network in pandapower format

    """

    time_start = time.perf_counter()
    eq_type = eq_type.lower()
    net = deepcopy(net)
    if not len(boundary_buses):
        raise ValueError("No boundary buses are given.")
    _check_network(net)
    logger.info(eq_type + " equivalent calculation started.")

    # --- determine interal buses, external buses, buses connected to boundary buses via
    #     bus-bus-switch and update boundary buses by external slack buses

    all_internal_buses, all_external_buses, boundary_buses_inclusive_bswitch, boundary_buses = \
        _determine_bus_groups(net, boundary_buses, internal_buses, show_computing_time)
    if not len(all_external_buses):
        logger.warning("There are no external buses so that no equivalent grid can be calculated.")
        return None
    return_internal &= bool(len(all_internal_buses))

    # --- ensure unique boundary bus names
    _ensure_unique_boundary_bus_names(net, boundary_buses_inclusive_bswitch)

    # --- create reference buses
    orig_slack_gens = add_ext_grids_to_boundaries(
        net, boundary_buses, adapt_va_degree, calc_volt_angles=calculate_voltage_angles,
        allow_net_change_for_convergence=allow_net_change_for_convergence, runpp_fct=runpp_fct,
        **kwargs)

    # --- replace ward and xward elements by internal elements (load, shunt, impedance, gen)
    ext_buses_with_ward = net.ward.bus[net.ward.bus.isin(all_external_buses)]
    ext_buses_with_xward = net.xward.bus[net.xward.bus.isin(all_external_buses)]
    if len(ext_buses_with_ward):
        logger.debug("ward elements of the external network are replaced by internal elements.")
        replace_ward_by_internal_elements(net, wards=ext_buses_with_ward.index)
    if len(ext_buses_with_xward):
        logger.debug("xward elements of the external network are replaced by internal elements.")
        replace_xward_by_internal_elements(net, xwards=ext_buses_with_xward.index)

    # --- switch from ward injection to ward addmittance if requested
    if eq_type in ["ward", "xward"] and ward_type == "ward_admittance":
        create_passive_external_net_for_ward_admittance(
            net, all_external_buses, boundary_buses, runpp_fct=runpp_fct, **kwargs)

    # --- rei calculations
    if eq_type == "rei":
        # --- create zero power balance network
        net_zpbn, net_internal, _ = _create_net_zpbn(
            net, boundary_buses, all_internal_buses,
            all_external_buses, calc_volt_angles=calculate_voltage_angles,
            runpp_fct=runpp_fct, **kwargs)

        # --- determine bus-lookups for the following calculation
        bus_lookups = _create_bus_lookups(
            net_zpbn, boundary_buses, all_internal_buses,
            all_external_buses, boundary_buses_inclusive_bswitch,
            show_computing_time)

        # --- calculate equivalent Ybus according to gaussian elimination
        Ybus_eq = _calculate_equivalent_Ybus(net_zpbn, bus_lookups,
                                             eq_type, show_computing_time,
                                             **kwargs)

        # --- calculate equivalent impedance and shunts
        shunt_params, impedance_params = \
            _calclate_equivalent_element_params(net_zpbn, Ybus_eq, bus_lookups,
                                                show_computing_time, **kwargs)

        # --- replace external network by equivalent elements
        _replace_ext_area_by_impedances_and_shunts(
            net_zpbn, bus_lookups, impedance_params, shunt_params,
            net_internal, return_internal, show_computing_time,
            calc_volt_angles=calculate_voltage_angles, runpp_fct=runpp_fct)
        net_eq = net_zpbn

    # --- ward and xward calculations
    elif eq_type in ["ward", "xward"]:
        net_internal, net_external = _get_internal_and_external_nets(
            net, boundary_buses, all_internal_buses, all_external_buses,
            calc_volt_angles=calculate_voltage_angles, runpp_fct=runpp_fct)

        # --- remove buses without power flow results in net_eq
        drop_buses(net_external, net_external.res_bus.index[net_external.res_bus.vm_pu.isnull()])

        # --- determine bus-lookups for the following calculation
        bus_lookups = _create_bus_lookups(
            net_external, boundary_buses, all_internal_buses,
            all_external_buses, boundary_buses_inclusive_bswitch)

        # --- cacluate equivalent Ybus accourding to gaussian elimination
        Ybus_eq = _calculate_equivalent_Ybus(net_external, bus_lookups,
                                             eq_type, show_computing_time,
                                             check_validity=False)

        if eq_type == "ward":
            # --- calculate equivalent impedance and wards
            ward_parameter_no_power, impedance_parameter = \
                _calculate_ward_and_impedance_parameters(Ybus_eq, bus_lookups,
                                                         show_computing_time)

            # --- replace external network by equivalent elements
            _replace_external_area_by_wards(net_external, bus_lookups,
                                            ward_parameter_no_power,
                                            impedance_parameter,
                                            ext_buses_with_xward,
                                            show_computing_time,
                                            calc_volt_angles=calculate_voltage_angles,
                                            runpp_fct=runpp_fct)
        else:  # eq_type == "xward"
            # --- calculate equivalent impedance and xwards
            xward_parameter_no_power, impedance_parameter = \
                _calculate_xward_and_impedance_parameters(net_external,
                                                          Ybus_eq,
                                                          bus_lookups,
                                                          show_computing_time)

            # --- replace external network by equivalent elements
            _replace_external_area_by_xwards(net_external, bus_lookups,
                                             xward_parameter_no_power,
                                             impedance_parameter,
                                             ext_buses_with_xward,
                                             show_computing_time,
                                             calc_volt_angles=calculate_voltage_angles,
                                             runpp_fct=runpp_fct)
        net_eq = net_external
    else:
        raise NotImplementedError(f"The {eq_type=} is unknown.")

    net_eq["bus_lookups"] = bus_lookups

    if return_internal:
        logger.debug("Merging of internal and equivalent network begins.")
        if len(kwargs.get("central_controller_types", [])):
            net_internal.controller.drop([idx for idx in net_internal.controller.index if any([
                isinstance(net_internal.controller.object.at[idx], central_controller_type) for
                central_controller_type in kwargs["central_controller_types"]])], inplace=True)
        net_eq = merge_internal_net_and_equivalent_external_net(
            net_eq, net_internal, show_computing_time=show_computing_time,
            calc_volt_angles=calculate_voltage_angles)
        if len(orig_slack_gens):
            net_eq.gen.loc[net_eq.gen.index.intersection(orig_slack_gens), "slack"] = True
        # run final power flow calculation
        net_eq = runpp_fct(net_eq, calculate_voltage_angles=calculate_voltage_angles)
    else:
        drop_assist_elms_by_creating_ext_net(net_eq)
        logger.debug("Only the equivalent net is returned.")

    # delete bus in poly_cost
    match_cost_functions_and_eq_net(net_eq, boundary_buses, eq_type)

    time_end = time.perf_counter()
    logger.info("%s equivalent finished in %.2f seconds." % (eq_type, time_end - time_start))

    if kwargs.get("add_group", True):
        # declare a group for the new equivalent
        ib_buses_after_merge, be_buses_after_merge = \
            _get_buses_after_merge(net_eq, net_internal, bus_lookups, return_internal)
        eq_elms = dict()
        for elm in ["bus", "gen", "impedance", "load", "sgen", "shunt",
                    "switch", "ward", "xward"]:
            if "ward" in elm:
                new_idx = net_eq[elm].index[net_eq[elm].name == "network_equivalent"].difference(
                    net[elm].index[net[elm].name == "network_equivalent"])
            else:
                names = net_eq[elm].name.astype(str)
                if elm in ["bus", "sgen", "gen", "load"]:
                    buses = net_eq.bus.index if elm == "bus" else net_eq[elm].bus
                    new_idx = net_eq[elm].index[names.str.contains("_integrated") |
                                                names.str.contains("_separate") &
                                                ~buses.isin(ib_buses_after_merge)]
                elif elm in ["impedance"]:
                    fr_buses = net_eq[elm].from_bus
                    to_buses = net_eq[elm].to_bus
                    new_idx = net_eq[elm].index[names.str.startswith("eq_%s" % elm) |
                                                (fr_buses.isin(be_buses_after_merge) &
                                                 to_buses.isin(be_buses_after_merge))]
                else:
                    buses = net_eq[elm].bus
                    new_idx = net_eq[elm].index[names.str.startswith("eq_%s" % elm) &
                                                buses.isin(be_buses_after_merge)]

                # don't include eq elements to this Group if these are already included to other
                # groups
                # ATTENTION: If there are eq elements (elements that fit to the above query of
                # new_idx) which already exist in net but are not included to other groups, they
                # will be considered here which is wrong. Furthermore, the indices may have changed
                # from net to net_eq, so that already existing groups with reference_columns == None
                # may fail their functionality
                new_idx = new_idx[~isin_group(net_eq, elm, new_idx)]

            if len(new_idx):
                eq_elms[elm] = list(new_idx)

        gr_idx = create_group_from_dict(net_eq, eq_elms, name=kwargs.get("group_name", eq_type))
        reference_column = kwargs.get("reference_column", None)
        if reference_column is not None:
            set_group_reference_column(net_eq, gr_idx, reference_column)

    return net_eq


def merge_internal_net_and_equivalent_external_net(
        net_eq, net_internal, fuse_bus_column="auto", show_computing_time=False, **kwargs):
    """
    Merges the internal network and the equivalent external network.
    It is expected that the boundaries occur in both, equivalent net and
    internal net. Therefore, the boundaries are first dropped in the
    internal net before merging.

    INPUT:
        **net_eq** - equivalent external area

        **net_internal** - internal area

    OPTIONAL:
        **fuse_bus_column**  (str, "auto) - the function expects boundary buses to be in net_eq and
        in net_internal. These duplicate buses get fused. To identify these buses, the given column is used. Option "auto" provides backward compatibility which is: use "name_equivalent" if
        existing and "name" otherwise

        **show_computing_time** (bool, False)

        ****kwargs** - key word arguments for pp.merge_nets()

    OUTPUT:
        **merged_net** - equivalent network within the internal area

    """
    kwargs = deepcopy(kwargs)
    t_start = time.perf_counter()
    net_internal["bus_lookups"] = net_eq["bus_lookups"]
    boundary_buses_inclusive_bswitch = net_eq.bus_lookups["boundary_buses_inclusive_bswitch"]

    # --- drop all branch elements between boundary buses in the internal net
    drop_internal_branch_elements(net_internal, boundary_buses_inclusive_bswitch)

    # --- drop bus elements attached to boundary buses in the internal net
    if kwargs.pop("drop_boundary_buses", True):
        drop_elements_at_buses(net_internal, boundary_buses_inclusive_bswitch, branch_elements=False)

    # --- merge equivalent external net and internal net
    merged_net = merge_nets(
        net_internal, net_eq, validate=kwargs.pop("validate", False),
        net2_reindex_log_level=kwargs.pop("net2_reindex_log_level", "debug"), **kwargs
    )

    # --- fuse or combine the boundary buses in external and internal nets
    if fuse_bus_column == "auto":
        if fuse_bus_column in merged_net.bus.columns:
            raise ValueError(
                f"{fuse_bus_column=} is ambiguous since the column 'auto' exists in net.bus")
        if "name_equivalent" in merged_net.bus.columns:
            fuse_bus_column = "name_equivalent"
        else:
            fuse_bus_column = "name"
    for bus in boundary_buses_inclusive_bswitch:
        try:
            name = merged_net.bus[fuse_bus_column].loc[bus]
        except:
            print(fuse_bus_column)
            print(merged_net.bus.columns)
            print()
        target_buses = merged_net.bus.index[merged_net.bus[fuse_bus_column] == name]
        if len(target_buses) != 2:
            raise ValueError(
                "The code expects all boundary buses to occur double. One because "
                "of net_eq and one because of net_internal. However target_buses is "
                "'%s'." % str(target_buses))
        fuse_buses(merged_net, target_buses[0], target_buses[1])

    # --- drop assist elements
    drop_assist_elms_by_creating_ext_net(merged_net)

    # --- drop repeated characteristic
    drop_repeated_characteristic(merged_net)

    # --- reindex buses named with "total" (done by REI)
    is_total_bus = merged_net.bus.name.astype(str).str.contains("total", na=False)
    if sum(is_total_bus):
        max_non_total_bus_idx = merged_net.bus[~is_total_bus].index.values.max()
        lookup = dict(zip(merged_net.bus.index[is_total_bus], range(
            max_non_total_bus_idx + 1, max_non_total_bus_idx + sum(is_total_bus) + 1)))
        reindex_buses(merged_net, lookup)

    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("'merge_int_and_eq_net' finished in %s seconds." %
                    round((t_end - t_start), 2))

    return merged_net


def drop_repeated_characteristic(net):
    for characteristic in ["trafo_characteristic_table", "shunt_characteristic_table",
                           "trafo_characteristic_spline", "shunt_characteristic_spline"]:
        if characteristic in net:
            # get the list of indices of duplicate rows
            repeated_idxs = net[characteristic][
                net[characteristic].fillna("NA_PLACEHOLDER").duplicated()].index.tolist()
            # drop repeated rows
            net[characteristic] = net[characteristic].drop(repeated_idxs)


def _determine_bus_groups(net, boundary_buses, internal_buses, show_computing_time=False):
    """
    Defines bus groups according to the given boundary buses and internal
    buses.

    INPUT:
        **net** - original network

        **boundary_buses** (iterable) - list of boundary bus indices

        **internal_buses** (iterable) - some of the internal bus indices

    OUTPUT:
        **all_internal_buses** (list) - list of all internal bus indices

        **all_external_buses** (list) - list of all external bus indices

        **boundary_buses_inclusive_bswitch** (list) - list of boundary buses
            and the connected buses via bus-bus-switch

        **boundary_buses** (list) - list of boundary bus indices (only
            supplied buses are considered)

    """
    # --- check internal and boundary buses
    t_start = time.perf_counter()
    if internal_buses is None:
        internal_buses = set()
    else:
        internal_buses = set(internal_buses)

    boundary_buses = set(boundary_buses)

    unsupplied_buses = set(net.res_bus.index[net.res_bus.vm_pu.isnull()])
    unsupplied_boundary_buses = boundary_buses & unsupplied_buses
    if len(unsupplied_boundary_buses):
        raise ValueError(
            "get_equivalent() do not allow unsupplied boundary " +
            "buses, these have no voltage results (possibly because power " +
            "flow results miss, the buses are isolated " +
            "or out of service): " + str(sorted(unsupplied_boundary_buses)) +
            ". Remove these buses from the boundary buses and try again.")

    if internal_buses & boundary_buses:
        logger.info("Some internal buses are also contained in the boundary buses, " +
                    "this could cause small inaccuracy.")

    # --- determine buses connected to boundary buses via bus-bus-switch
    boundary_buses_inclusive_bswitch = set()
    mg_sw = create_nxgraph(net, respect_switches=True, include_lines=False, include_impedances=False,
                           include_tcsc=False, include_trafos=False, include_trafo3ws=False)
    for bbus in boundary_buses:
        boundary_buses_inclusive_bswitch |= set(connected_component(mg_sw, bbus))
    if len(boundary_buses_inclusive_bswitch) > len(boundary_buses):
        logger.info("There are some buses connected to the boundary buses via " +
                    "bus-bus-switches. They could be the nodes on the same bus bar " +
                    "of the boundary buses. It is suggested to consider all these " +
                    "buses (the connected buses and the given boundary buses) " +
                    "as the boundary. They are: %s." % boundary_buses_inclusive_bswitch)

    # --- determine all internal buses
    all_internal_buses = set()
    if internal_buses is None:
        internal_buses = set()
    else:
        internal_buses = set(internal_buses)
        mg = create_nxgraph(net)
        cc = connected_components(mg, notravbuses=boundary_buses_inclusive_bswitch)
        while True:
            try:
                buses = next(cc)
            except StopIteration:
                break
            if len(buses & set(internal_buses)):
                all_internal_buses |= buses
        all_internal_buses -= boundary_buses_inclusive_bswitch

    # --- determine all external buses
    all_external_buses = set(net.bus.index) - unsupplied_buses - \
                         all_internal_buses - boundary_buses_inclusive_bswitch

    # --- move all slack buses from external net to boundary if no slack is
    #     in boundary or internal
    slack_buses = set(net.ext_grid.bus[net.ext_grid.in_service]) | \
                  set(net.gen.bus[net.gen.in_service & net.gen.slack])
    if len(all_internal_buses) and not len((
                                                   all_internal_buses | boundary_buses_inclusive_bswitch) &
                                           slack_buses):
        if not len(slack_buses):
            raise ValueError("There is no active slack in the net.")
        for bbus in slack_buses & all_external_buses:
            boundary_buses |= {bbus}
            bbus_bswitch = set(connected_component(mg_sw, bbus))
            boundary_buses_inclusive_bswitch |= bbus_bswitch
            all_external_buses -= {bbus} | bbus_bswitch

    # --- function endings
    _check_bus_groups(all_internal_buses, all_external_buses, internal_buses,
                      boundary_buses)
    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("\"determine_bus_groups\" finished in %s seconds." %
                    round((t_end - t_start), 2))

    return sorted(all_internal_buses), sorted(all_external_buses), \
        sorted(boundary_buses_inclusive_bswitch), sorted(boundary_buses)


def _check_bus_groups(all_internal_buses, all_external_buses, internal_buses, boundary_buses):
    """
    Checks the plausibility of the bus groups.
    """
    missing_internals = internal_buses - all_internal_buses
    if len(missing_internals) and not (missing_internals & boundary_buses):
        raise ValueError("These internal buses miss in 'all_internal_buses': " + str(sorted(
            missing_internals)))
    in_and_extern_buses = all_internal_buses & all_external_buses
    if len(in_and_extern_buses):
        raise ValueError("These buses are in 'all_internal_buses' and 'all_external_buses': " + str(
            sorted(in_and_extern_buses)))


def _get_buses_after_merge(net_eq, net_internal, bus_lookups, return_internal):
    """
    Finds bus groups according to the new index after merge
    """
    if return_internal:
        bus_list = net_eq.bus.index.tolist()
        bus_list_old = net_internal.bus.index.tolist()
        ib_buses_after_merge = bus_list[:sum(~net_internal.bus.name.str.contains("assist_", na=False))]
        i_pos = [bus_list_old.index(i) for i in bus_lookups["origin_all_internal_buses"]]
        be_buses_after_merge = [x for i, x in enumerate(bus_list) if i not in i_pos]
    else:
        ib_buses_after_merge = bus_lookups["bus_lookup_pd"]["b_area_buses"]
        be_buses_after_merge = bus_lookups["bus_lookup_pd"]["b_area_buses"] + \
                               bus_lookups["bus_lookup_pd"]["e_area_buses"]
    return ib_buses_after_merge, be_buses_after_merge


if __name__ == "__main__":
    """ --- quick test --- """
    # pp.logger.setLevel(logging.ERROR)
    # logger.setLevel(logging.DEBUG)
    from pandapower.networks import case9

    net = case9()
    net.ext_grid.vm_pu = 1.04
    net.gen.vm_pu[0] = 1.025
    net.gen.vm_pu[1] = 1.025

    net.poly_cost = net.poly_cost.drop(net.poly_cost.index)
    net.pwl_cost = net.pwl_cost.drop(net.pwl_cost.index)
    # pp.replace_gen_by_sgen(net)
    # net.sn_mva = 109.00
    boundary_buses = [4, 8]
    internal_buses = [0]
    return_internal = True
    show_computing_time = False
    runpp(net, calculate_voltage_angles=True)
    net_org = deepcopy(net)
    eq_type = "rei"
    net_eq = get_equivalent(net, eq_type, boundary_buses,
                            internal_buses,
                            return_internal=return_internal,
                            show_computing_time=False,
                            calculate_voltage_angles=True)
    print(net.res_bus)
    # print(net_eq.res_bus.loc[[0,3]])
    # print(net_eq.ward.loc[0])

    # net_eq.sn_mva = 10
    # pp.runpp(net_eq, calculate_voltage_angles=True)
    # print(net_eq.res_bus.loc[[0,3]])



import logging
import time

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.externalNetworkInjectionsCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class ExternalNetworkInjectionsCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_external_network_injections_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting ExternalNetworkInjections.")

        eqssh_eni = self._prepare_external_network_injections_cim16()

        # choose the slack
        eni_ref_prio_min = eqssh_eni.loc[(eqssh_eni['enabled']) & (eqssh_eni['slack_weight'] > 0), 'slack_weight'].min()
        # check if the slack is a SynchronousMachine
        sync_machines = self.cimConverter.merge_eq_ssh_profile('SynchronousMachine')
        sync_machines = self.get_voltage_from_controllers(sync_machines)

        sync_ref_prio_min = sync_machines.loc[
            (sync_machines['referencePriority'] > 0) & (sync_machines['enabled']), 'referencePriority'].min()
        if pd.isna(eni_ref_prio_min):
            ref_prio_min = sync_ref_prio_min
        elif pd.isna(sync_ref_prio_min):
            ref_prio_min = eni_ref_prio_min
        else:
            ref_prio_min = min(eni_ref_prio_min, sync_ref_prio_min)

        eni_slacks = eqssh_eni.loc[(eqssh_eni['slack_weight'] == ref_prio_min) & (eqssh_eni['controllable'])]
        eni_gens = eqssh_eni.loc[(eqssh_eni['slack_weight'] != ref_prio_min) & (eqssh_eni['controllable'])]
        eni_sgens = eqssh_eni.loc[~eqssh_eni['controllable']]

        # create reactive_capability_curve flag
        if 'reactive_capability_curve' not in eni_gens.columns:
            eni_gens['reactive_capability_curve'] = False
        # create reactive_capability_curve flag
        if 'reactive_capability_curve' not in eni_sgens.columns:
            eni_sgens['reactive_capability_curve'] = False
        # create reactive_capability_curve flag
        if 'reactive_capability_curve' not in eni_slacks.columns:
            eni_slacks['reactive_capability_curve'] = False

        self.cimConverter.copy_to_pp('ext_grid', eni_slacks)
        self.cimConverter.copy_to_pp('gen', eni_gens)
        self.cimConverter.copy_to_pp('sgen', eni_sgens)

        self.logger.info("Created %s external networks, %s generators and %s static generators in %ss" %
                         (eni_slacks.index.size, eni_gens.index.size, eni_sgens.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s external networks, %s generators and %s static generators from "
                    "ExternalNetworkInjections in %ss" %
                    (eni_slacks.index.size, eni_gens.index.size, eni_sgens.index.size, time.time() - time_start)))

    def _prepare_external_network_injections_cim16(self) -> pd.DataFrame:
        if 'sc' in self.cimConverter.cim.keys():
            eni = self.cimConverter.merge_eq_other_profiles(['ssh', 'sc'], 'ExternalNetworkInjection',
                                                        add_cim_type_column=True)
        else:
            eni = self.cimConverter.merge_eq_ssh_profile('ExternalNetworkInjection', add_cim_type_column=True)

        # merge with buses
        eni = pd.merge(eni, self.cimConverter.bus_merge, how='left', on='rdfId')

        # get the voltage from controllers
        eni = self.get_voltage_from_controllers(eni)

        # get slack voltage and angle from SV profile
        eni = pd.merge(eni, self.cimConverter.net.bus[['vn_kv', sc['ct']]],
                       how='left', left_on='index_bus', right_index=True)
        eni = pd.merge(eni, self.cimConverter.cim['sv']['SvVoltage'][['TopologicalNode', 'v', 'angle']],
                       how='left', left_on=sc['ct'], right_on='TopologicalNode')
        eni['controlEnabled'] = eni['controlEnabled'] & eni['enabled']
        eni['vm_pu'] = eni['targetValue'] / eni['vn_kv']  # voltage from regulation
        eni['vm_pu'] = eni['vm_pu'].fillna(eni['v'] / eni['vn_kv'])  # voltage from measurement
        eni['vm_pu'] = eni['vm_pu'].fillna(1.)  # default voltage
        eni['angle'] = eni['angle'].fillna(0.)  # default angle
        eni['ratedU'] = eni['targetValue'][:]  # targetValue in kV
        eni['ratedU'] = eni['ratedU'].fillna(eni['v'])  # v in kV
        eni['ratedU'] = eni['ratedU'].fillna(eni['vn_kv'])
        eni['s_sc_max_mva'] = 3 ** .5 * eni['ratedU'] * (eni['maxInitialSymShCCurrent'] / 1e3)
        eni['s_sc_min_mva'] = 3 ** .5 * eni['ratedU'] * (eni['minInitialSymShCCurrent'] / 1e3)
        # get the substations
        eni = pd.merge(eni,
                       self.cimConverter.net.bus[[sc['o_id'], 'zone']].rename({sc['o_id']: 'b_id'}, axis=1),
                       how='left', left_on='ConnectivityNode', right_on='b_id')
        
        eni['referencePriority'] = eni['referencePriority'].astype(float)
        eni['p'] = -eni['p']
        eni['q'] = -eni['q']
        eni['x0x_max'] = ((eni['maxR1ToX1Ratio'] + 1j) /
                          (eni['maxR0ToX0Ratio'] + 1j)).abs() * eni['maxZ0ToZ1Ratio']

        if 'inService' in eni.columns:
            eni['connected'] = eni['connected'] & eni['inService']

        eni = eni.rename(columns={'rdfId': sc['o_id'], 'rdfId_Terminal': sc['t'], 'zone': sc['sub'],
                                  'angle': 'va_degree', 'index_bus': 'bus', 'connected': 'in_service',
                                  'minP': 'min_p_mw', 'maxP': 'max_p_mw', 'minQ': 'min_q_mvar', 'maxQ': 'max_q_mvar',
                                  'p': 'p_mw', 'q': 'q_mvar', 'controlEnabled': 'controllable',
                                  'maxR1ToX1Ratio': 'rx_max', 'minR1ToX1Ratio': 'rx_min', 'maxR0ToX0Ratio': 'r0x0_max',
                                  'referencePriority': 'slack_weight'})
        eni['scaling'] = 1.
        eni['type'] = None
        eni['slack'] = False

        return eni

    def get_voltage_from_controllers(self, eqssh_eni):
        regulation_controllers = self.cimConverter.merge_eq_ssh_profile('RegulatingControl')
        regulation_controllers = regulation_controllers.loc[regulation_controllers['mode'] == 'voltage']
        regulation_controllers = regulation_controllers[['rdfId', 'targetValue', 'enabled']]
        regulation_controllers = regulation_controllers.rename(columns={'rdfId': 'RegulatingControl'})
        eqssh_eni = pd.merge(eqssh_eni, regulation_controllers, how='left', on='RegulatingControl')
        return eqssh_eni



import copy

from pandapower.auxiliary import get_free_id
from pandapower.io_utils import JSONSerializableClass

try:
    import pandaplan.core.pplog as pplog
except ImportError:
    import logging as pplog

logger = pplog.getLogger(__name__)


class ProtectionDevice(JSONSerializableClass):
    """
    Base Class of protection devices found in protection module
    """
    def __init__(self, net, index=None, in_service=True, overwrite=False, **kwargs):
        super().__init__()
        # add oneself to net, creating the ['controller'] DataFrame, if necessary
        if index is None and "protection" in net.keys():
            index = get_free_id(net.protection)
        fill_dict = {"in_service": in_service}
        self.index = super().add_to_net(net=net, element='protection', index=index, overwrite=overwrite,
                                        fill_dict=fill_dict, preserve_dtypes=True)

    def reset_device(self):
        pass

    def has_tripped(self):
        logger.warning("Method has not been overwritten")
        return False

    def protection_function(self, net, scenario):
        logger.warning("Method has not been overwritten")

    def __repr__(self):
        rep = "This " + self.__class__.__name__ + " has the following parameters: \n"

        for member in ["index", "json_excludes"]:
            rep += ("\n" + member + ": ").ljust(20)
            d = locals()
            exec('value = self.' + member, d)
            rep += str(d['value'])

        return rep

    def __str__(self):
        s = self.__class__.__name__
        return s

    def __getstate__(self):
        state = copy.deepcopy(self.__dict__)

        for attr in self.json_excludes:
            try:
                del state[attr]
            except:
                continue

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)


if __name__ == "__main__":
    from pandapower.create import create_empty_network

    net = create_empty_network()
    ProtectionDevice(net)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import copy
import importlib
import json
import numbers
import os
import io
import sys
import types
import weakref
from ast import literal_eval
from functools import partial
from inspect import isclass, _findclass
from warnings import warn
import numpy as np
import pandas.errors
from deepdiff.diff import DeepDiff
from packaging.version import Version
from pandapower._version import __version__
import networkx
import numpy
import geojson
import pandas as pd
from networkx.readwrite import json_graph
from numpy import ndarray, generic, equal, isnan, allclose, any as anynp

try:
    import psycopg2
    import psycopg2.errors
    import psycopg2.extras
    PSYCOPG2_INSTALLED = True
except ImportError:
    psycopg2 = None
    PSYCOPG2_INSTALLED = False
try:
    from pandas.testing import assert_series_equal, assert_frame_equal
except ImportError:
    from pandas.util.testing import assert_series_equal, assert_frame_equal
try:
    from cryptography.fernet import Fernet
    cryptography_INSTALLED = True
except ImportError:
    cryptography_INSTALLED = False
try:
    import hashlib
    hashlib_INSTALLED = True
except ImportError:
    hashlib_INSTALLED = False
try:
    import base64
    base64_INSTALLED = True
except ImportError:
    base64_INSTALLED = False
try:
    import zlib
    zlib_INSTALLED = True
except ImportError:
    zlib_INSTALLED = False

from pandapower.auxiliary import pandapowerNet, get_free_id, soft_dependency_error, _preserve_dtypes
from pandapower.create import create_empty_network

from functools import singledispatch

try:
    import geopandas

    GEOPANDAS_INSTALLED = True
except ImportError:
    GEOPANDAS_INSTALLED = False

try:
    import shapely.geometry

    SHAPELY_INSTALLED = True
except (ImportError, OSError):
    SHAPELY_INSTALLED = False

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def coords_to_df(value, geotype="line"):
    columns = ["x", "y", "coords"] if geotype == "bus" else ["coords"]
    geo = pd.DataFrame(columns=columns, index=value.index)
    if "coords" in value.columns and any(~value.coords.isnull()):
        k = max(len(v) for v in value.coords.values)
        v = numpy.empty((len(value), k * 2))
        v.fill(numpy.nan)
        for i, idx in enumerate(value.index):
            # get coords and convert them to x1, y1, x2, y2...
            coords = value.at[idx, 'coords']
            if coords is None:
                continue
            v[i, :len(coords) * 2] = numpy.array(coords).flatten()
        geo = pd.DataFrame(v, index=value.index)
        geo.columns = ["%s%i" % (w, i) for i in range(k) for w in "xy"]
    if geotype == "bus":
        geo["x"] = value["x"].values
        geo["y"] = value["y"].values
    return geo


def to_dict_of_dfs(net, include_results=False, include_std_types=True, include_parameters=True,
                   include_empty_tables=True):
    dodfs = dict()
    dtypes = []
    parameters = dict()  # pd.DataFrame(columns=["parameter"])
    for item, value in net.items():
        # dont save internal variables and results (if not explicitely specified)
        if item.startswith("_") or (item.startswith("res") and not include_results):
            continue
        elif item == "std_types":
            if not include_std_types:
                continue
            for t in net.std_types.keys():  # which are ["line", "trafo", "trafo3w", "fuse"]
                if net.std_types[t]:  # avoid empty excel sheets for std_types if empty
                    type_df = pd.DataFrame(net.std_types[t]).T
                    if t == "fuse":
                        for c in type_df.columns:
                            type_df[c] = type_df[c].apply(lambda x: str(x) if isinstance(x, list) else x)
                    dodfs["%s_std_types" % t] = type_df
            continue
        elif item == "profiles":
            for t in net.profiles.keys():  # which could be e.g. "sgen", "gen", "load", ...
                if net.profiles[t].shape[0]:  # avoid empty excel sheets for std_types if empty
                    dodfs["%s_profiles" % t] = pd.DataFrame(net.profiles[t])
            continue
        elif item == "user_pf_options":
            if len(value) > 0:
                dodfs["user_pf_options"] = pd.DataFrame(value, index=[0])
            continue
        elif isinstance(value, (int, float, bool, str, numbers.Number)):
            # attributes of primitive types are just stored in a DataFrame "parameters"
            parameters[item] = net[item]
            continue
        elif not isinstance(value, pd.DataFrame):
            logger.warning("Could not serialize net.%s" % item)
            continue

        # value is pandas DataFrame
        if not include_empty_tables and value.empty:
            continue

        if item == "bus_geodata":
            geo = coords_to_df(value, geotype="bus")
            if GEOPANDAS_INSTALLED and isinstance(value, geopandas.GeoDataFrame):
                geo["geometry"] = [s.to_wkt() for s in net.bus_geodata.geometry.values]
            dodfs[item] = geo
        elif item == "line_geodata":
            geo = coords_to_df(value, geotype="line")
            if GEOPANDAS_INSTALLED and isinstance(value, geopandas.GeoDataFrame):
                geo["geometry"] = [s.to_wkt() for s in net.line_geodata.geometry.values]
            dodfs[item] = geo
        elif "object" in value.columns:
            columns = [c for c in value.columns if c != "object"]
            tab = value[columns].copy()
            tab["object"] = value["object"].apply(lambda x: json.dumps(x, cls=PPJSONEncoder, indent=2))
            tab = tab[value.columns]
            if "recycle" in tab.columns:
                tab["recycle"] = tab["recycle"].apply(json.dumps)
            dodfs[item] = tab
        elif "geo" in value.columns:
            columns = [c for c in value.columns if c != "geo"]
            tab = value[columns].copy()
            tab["geo"] = value["geo"].apply(lambda x: json.dumps(x, cls=PPJSONEncoder, indent=2))
            tab = tab[value.columns]
            dodfs[item] = tab
        else:
            dodfs[item] = value
        # save dtypes
        for column, dtype in value.dtypes.items():
            dtypes.append((item, column, str(dtype)))
    dodfs["dtypes"] = pd.DataFrame(dtypes, columns=["element", "column", "dtype"])
    if include_parameters and len(parameters) > 0:
        dodfs["parameters"] = pd.DataFrame(parameters, index=[0])
    return dodfs


def dicts_to_pandas(json_dict):
    warn("This function is deprecated and will be removed in a future release.\r\n"
         "Please resave your grid using the current pandapower version.", DeprecationWarning)
    pd_dict = dict()
    for k in sorted(json_dict.keys()):
        if isinstance(json_dict[k], dict):
            pd_dict[k] = pd.DataFrame.from_dict(json_dict[k], orient="columns")
            if pd_dict[k].shape[0] == 0:  # skip empty dataframes
                continue
            if pd_dict[k].index[0].isdigit():
                pd_dict[k].set_index(pd_dict[k].index.astype(numpy.int64), inplace=True)
        else:
            raise UserWarning("The network is an old version or corrupt. "
                              "Try to use the old load function")
    return pd_dict


def df_to_coords(net, item, table):
    # converts dataframe to coords in net
    num_points = len(table.columns) // 2
    net[item] = pd.DataFrame(index=table.index, columns=net[item].columns)
    if item == "bus_geodata":
        num_points -= 1
        net[item].loc[:, ['x', 'y']] = table.loc[:, ['x', 'y']]

    for i in table.index:
        coords = table.loc[i]
        # for i, coords in table.iterrows():
        coord = [(coords["x%u" % nr], coords["y%u" % nr]) for nr in range(num_points)
                 if pd.notnull(coords["x%u" % nr])]
        if len(coord):
            net[item].loc[i, "coords"] = coord


def from_dict_of_dfs(dodfs, net=None):
    if net is None:
        net = create_empty_network()
    for item, table in dodfs.items():
        if item == "dtypes":
            continue
        elif item == "parameters":
            for c in dodfs["parameters"].columns:
                net[c] = dodfs["parameters"].at[0, c]
                if c == "name" and pd.isnull(net[c]):
                    net[c] = ''
            continue
        elif item in ["line_geodata", "bus_geodata"]:
            pass
        #    table = table.rename_axis(net[item].index.name)
        #    df_to_coords(net, item, table)
        elif item.endswith("_std_types"):
            # when loaded from Excel, the lists in the DataFrame cells are strings -> we want to convert them back
            # to lists here. There is probably a better way to deal with it.
            if item.startswith("fuse"):
                for c in table.columns:
                    table[c] = table[c].apply(lambda x: json.loads(x) if isinstance(x, str) and x.startswith("[") else x)
            net["std_types"][item[:-10]] = table.T.to_dict()
            continue  # don't go into try..except
        elif item.endswith("_profiles"):
            if "profiles" not in net.keys():
                net["profiles"] = dict()
            table = table.rename_axis(None)
            net["profiles"][item[:-9]] = table
            continue  # don't go into try..except
        elif item == "user_pf_options":
            net['user_pf_options'] = {c: v for c, v in zip(table.columns, table.values[0])}
            continue  # don't go into try..except
        else:
            for json_column in ("object", "recycle"):
                if json_column in table.columns:
                    table[json_column] = table[json_column].apply(
                        lambda x: json.loads(x, cls=PPJSONDecoder))
            if not isinstance(table.index, pd.MultiIndex):
                table = table.rename_axis(net[item].index.name)
            net[item] = table
            # convert geodata to geojson
            if item in ["bus", "line"]:
                if "geo" in table.columns:
                    table.geo = table.geo.apply(
                        lambda x: geojson.loads(x, cls=PPJSONDecoder) if pd.notna(x) else x
                    )
        # set the index to be Int
        try:
            net[item].set_index(net[item].index.astype(np.int64), inplace=True)
        except (TypeError, ValueError):
            # TypeError or ValueError: if not int index (e.g. str)
            pass
    if "dtypes" in dodfs:
        restore_all_dtypes(net, dodfs["dtypes"])
    elif "dtypes" in net:
        restore_all_dtypes(net, net["dtypes"])
    return net


def restore_all_dtypes(net, dtypes):
    for _, v in dtypes.iterrows():
        try:
            if v["dtype"] == "object":
                c = net[v.element][v.column]
                net[v.element][v.column] = numpy.where(c.isnull(), None, c)
                # net[v.element][v.column] = net[v.element][v.column].fillna(value=None)
            net[v.element][v.column] = net[v.element][v.column].astype(v["dtype"])
        except KeyError:
            pass
        except pandas.errors.IntCastingNaNError:
            logger.info(f"could not convert dtype in {v.element}: {v.column}")


def to_dict_with_coord_transform(net, point_geo_columns, line_geo_columns):
    save_net = dict()
    for key, item in net.items():
        if hasattr(item, "columns") and "geometry" in item.columns:
            # we convert shapely-objects to primitive data-types on a deepcopy
            item = copy.deepcopy(item)
            if key in point_geo_columns and not isinstance(item.geometry.values[0], tuple):
                item["geometry"] = item.geometry.apply(lambda x: (x.x, x.y))
            elif key in line_geo_columns and not isinstance(item.geometry.values[0], list):
                item["geometry"] = item.geometry.apply(lambda x: list(x.coords))

        save_net[key] = {"DF": item.to_dict("split"),
                         "dtypes": {col: dt for col, dt in zip(item.columns, item.dtypes)}} \
            if isinstance(item, pd.DataFrame) else item
    return save_net


def get_raw_data_from_pickle(filename):
    def read(f):
        return pd.read_pickle(f)

    if hasattr(filename, 'read'):
        net = read(filename)
    elif not os.path.isfile(filename):
        raise UserWarning("File %s does not exist!!" % filename)
    else:
        with open(filename, "rb") as f:
            net = read(f)
    return net


def transform_net_with_df_and_geo(net, point_geo_columns, line_geo_columns):
    try:
        epsg = net.gis_epsg_code
    except AttributeError:
        epsg = None

    for key, item in net.items():
        if isinstance(item, dict) and "DF" in item:
            df_dict = item["DF"]
            if "columns" in df_dict:
                # make sure the index is Int
                try:
                    df_index = pd.Index(df_dict['index'], dtype=numpy.int64)
                except TypeError:
                    df_index = df_dict['index']
                if GEOPANDAS_INSTALLED and "geometry" in df_dict["columns"] \
                        and epsg is not None:
                    # convert primitive data-types to shapely-objects
                    if key in point_geo_columns:
                        data = {"x": [row[0] for row in df_dict["data"]],
                                "y": [row[1] for row in df_dict["data"]]}
                        geo = [shapely.geometry.Point(row[2][0], row[2][1])
                               for row in df_dict["data"]]
                    elif key in line_geo_columns:
                        data = {"coords": [row[0] for row in df_dict["data"]]}
                        geo = [shapely.geometry.LineString(row[1]) for row in df_dict["data"]]

                    net[key] = geopandas.GeoDataFrame(data, crs=f"epsg:{epsg}", geometry=geo,
                                                      index=df_index)
                else:
                    net[key] = pd.DataFrame(columns=df_dict["columns"], index=df_index,
                                            data=df_dict["data"])
            else:
                net[key] = pd.DataFrame.from_dict(df_dict)
                if "columns" in item:
                    net[key] = net[key].reindex(item["columns"], axis=1)

            if "dtypes" in item:
                if "columns" in df_dict and "geometry" in df_dict["columns"]:
                    pass
                else:
                    try:
                        # only works with pandas 0.19 or newer
                        net[key] = net[key].astype(item["dtypes"])
                    except:
                        # works with pandas <0.19
                        for column in net[key].columns:
                            net[key][column] = net[key][column].astype(item["dtypes"][column])


def isinstance_partial(obj, cls):
    # this function shall make sure that for the given classes, no default string functions are
    # used, but the registered ones (to_serializable registry)
    if isinstance(obj, (pandapowerNet, tuple, numpy.floating)):
        return False
    return isinstance(obj, cls)


def check_net_version(net):
    if Version(net["format_version"]) > Version(__version__):
        logger.warning(f"pandapowerNet-version ({net['format_version']}) is newer than your "
                       f"pandapower version ({__version__}). Please update pandapower "
                       "`pip install --upgrade pandapower`.")


class PPJSONEncoder(json.JSONEncoder):
    def __init__(self, isinstance_func=isinstance_partial, **kwargs):
        super(PPJSONEncoder, self).__init__(**kwargs)
        self.isinstance_func = isinstance_func

    def iterencode(self, o, _one_shot=False):
        """Encode the given object and yield each string
        representation as available.

        For example::

            for chunk in JSONEncoder().iterencode(bigobject):
                mysocket.write(chunk)

        """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = json.encoder.encode_basestring_ascii
        else:
            _encoder = json.encoder.encode_basestring

        def floatstr(o, allow_nan=self.allow_nan, _repr=float.__repr__, _inf=json.encoder.INFINITY,
                     _neginf=-json.encoder.INFINITY):
            # Check for specials.  Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on the
            # internals.

            if o != o:
                text = 'NaN'
            elif o == _inf:
                text = 'Infinity'
            elif o == _neginf:
                text = '-Infinity'
            else:
                return _repr(o)

            if not allow_nan:
                raise ValueError(
                    "Out of range float values are not JSON compliant: " + repr(o))

            return text

        _iterencode = json.encoder._make_iterencode(
            markers, self.default, _encoder, self.indent, floatstr,
            self.key_separator, self.item_separator, self.sort_keys,
            self.skipkeys, _one_shot, isinstance=self.isinstance_func)
        return _iterencode(o, 0)

    def default(self, o):
        try:
            s = to_serializable(o)
        except TypeError:
            # Let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, o)
        else:
            return s


class FromSerializable:
    def __init__(self):
        self.class_name = 'class_name'
        self.module_name = 'module_name'
        self.registry = {}

    def __get__(self, instance, owner):
        if instance is None:
            return self
        class_module = getattr(instance, self.class_name), getattr(instance, self.module_name)
        if class_module not in self.registry:
            _class = (class_module[0], '')
            _module = ('', class_module[1])
            if (_class in self.registry) and (_module in self.registry):
                logger.error('the saved object %s is ambiguous. There are at least two possibilites'
                             ' to decode the object' % class_module)
            elif _class in self.registry:
                class_module = _class
            elif _module in self.registry:
                class_module = _module
            else:
                class_module = ('', '')
        method = self.registry[class_module]
        return method.__get__(instance, owner)

    def register(self, class_name='', module_name=''):
        def decorator(method):
            self.registry[(class_name, module_name)] = method
            return method

        return decorator


class FromSerializableRegistry():
    from_serializable = FromSerializable()
    class_name = ''
    module_name = ''

    def __init__(self, obj, d, pp_hook_funct, ignore_unknown_objects=False):
        self.obj = obj
        self.d = d
        self.pp_hook = pp_hook_funct
        self.ignore_unknown_objects = ignore_unknown_objects

    @from_serializable.register(class_name='Series', module_name='pandas.core.series')
    def Series(self):
        is_multiindex = self.d.pop('is_multiindex', False)
        index_name = self.d.pop('index_name', None)
        index_names = self.d.pop('index_names', None)
        ser = pd.read_json(io.StringIO(self.obj), precise_float=True, **self.d)

        # restore index name and Multiindex
        if index_name is not None:
            ser.index.name = index_name
        if is_multiindex:
            try:
                if len(ser) == 0:
                    ser.index = pd.MultiIndex.from_tuples([], names=index_names, dtype=np.int64)
                else:
                    ser.index = pd.MultiIndex.from_tuples(pd.Series(ser.index).apply(literal_eval).tolist())
            except:
                logger.warning("Converting index to multiindex failed.")
            else:
                if index_names is not None:
                    ser.index.names = index_names

        return ser

    @from_serializable.register(class_name='DataFrame', module_name='pandas.core.frame')
    def DataFrame(self):
        is_multiindex = self.d.pop('is_multiindex', False)
        is_multicolumn = self.d.pop('is_multicolumn', False)
        index_name = self.d.pop('index_name', None)
        index_names = self.d.pop('index_names', None)
        column_name = self.d.pop('column_name', None)
        column_names = self.d.pop('column_names', None)

        obj = self.obj
        if type(obj) == str and (not os.path.isabs(obj) or not obj.endswith('.json')):
            obj = io.StringIO(obj)

        df = pd.read_json(obj, precise_float=True, convert_axes=False, **self.d)

        if not df.shape[0] or self.d.get("orient", False) == "columns":
            try:
                df.set_index(df.index.astype(numpy.int64), inplace=True)
            except (ValueError, TypeError, AttributeError):
                logger.debug("failed setting index to int")
        if self.d.get("orient", False) == "columns":
            try:
                df.columns = df.columns.astype(numpy.int64)
            except (ValueError, TypeError, AttributeError):
                logger.debug("failed setting columns to int")

        # restore index name and Multiindex
        if index_name is not None:
            df.index.name = index_name
        if column_name is not None:
            df.columns.name = column_name
        if is_multiindex:
            try:
                if len(df) == 0:
                    df.index = pd.MultiIndex.from_frame(pd.DataFrame(columns=index_names, dtype=np.int64))
                else:
                    df.index = pd.MultiIndex.from_tuples(pd.Series(df.index).apply(literal_eval).tolist())
                # slower alternative code:
                # df.index = pd.MultiIndex.from_tuples([literal_eval(idx) for idx in df.index])
            except:
                logger.warning("Converting index to multiindex failed.")
            else:
                if index_names is not None:
                    df.index.names = index_names
        if is_multicolumn:
            try:
                if len(df) == 0:
                    df.columns = pd.MultiIndex.from_frame(pd.DataFrame(columns=column_names, dtype=np.int64))
                else:
                    df.columns = pd.MultiIndex.from_tuples(pd.Series(df.columns).apply(literal_eval).tolist())
            except:
                logger.warning("Converting columns to multiindex failed.")
            else:
                if column_names is not None:
                    df.columns.names = column_names

        if 'geo' in df.columns:
            df['geo'] = df['geo'].dropna().apply(json.dumps).apply(geojson.loads)

        df_obj = df.select_dtypes(include=['object'])
        for col in df_obj:
            df[col] = df[col].apply(partial(
                self.pp_hook, ignore_unknown_objects=self.ignore_unknown_objects
            ))
            df[col] = df[col].astype(dtype = 'object')
            df.loc[pd.isnull(df[col]), col] = None
        return df

    @from_serializable.register(class_name='pandapowerNet', module_name='pandapower.auxiliary')#,
                                # empty_dict_like_object=None)
    def pandapowerNet(self):
        if isinstance(self.obj, str):  # backwards compatibility
            from pandapower import from_json_string
            return from_json_string(self.obj)
        else:
            if self.empty_dict_like_object is None:
                net = create_empty_network()
            else:
                net = self.empty_dict_like_object
            net.update(self.obj)
            return net

    @from_serializable.register(class_name="MultiGraph", module_name="networkx")
    def networkx(self):
        mg = json_graph.adjacency_graph(self.obj, attrs={'id': 'json_id', 'key': 'json_key'})
        edges = list()
        for (n1, n2, e) in mg.edges:
            attr = {k: v for k, v in mg.get_edge_data(n1, n2, key=e).items() if
                    k not in ("json_id", "json_key")}
            attr["key"] = e
            edges.append((n1, n2, attr))
        mg.clear_edges()
        for n1, n2, ed in edges:
            mg.add_edge(n1, n2, **ed)
        return mg

    @from_serializable.register(class_name="method")
    def method(self):
        logger.warning('deserializing of method not implemented')
        # class_ = getattr(module, obj) # doesn't work
        return self.obj

    @from_serializable.register(class_name='function')
    def function(self):
        module = importlib.import_module(self.module_name)
        if not hasattr(module, self.obj):  # in case a function is a lambda or is not defined
            raise UserWarning('Could not find the definition of the function %s in the module %s' %
                              (self.obj, module.__name__))
        class_ = getattr(module, self.obj)  # works
        return class_

    @from_serializable.register()
    def rest(self):
        try:
            module = importlib.import_module(self.module_name)
        except ModuleNotFoundError as e:
            if self.ignore_unknown_objects:
                warn(f"Module {self.module_name} not found. Returning object as is.")
                return json.loads(self.obj)
            else:
                raise e
        try:
            class_ = getattr(module, self.class_name)
        except AttributeError as e:
            if self.ignore_unknown_objects:
                warn(f"Class {self.class_name} not found in module {self.module_name}. Returning object as is.")
                return json.loads(self.obj)
            else:
                raise e
        if isclass(class_) and issubclass(class_, JSONSerializableClass):
            if isinstance(self.obj, str):
                self.obj = json.loads(self.obj, cls=PPJSONDecoder,
                                      object_hook=partial(
                                          pp_hook, ignore_unknown_objects=self.ignore_unknown_objects
                                      ))
                # backwards compatibility
            if "net" in self.obj:
                del self.obj["net"]
            return class_.from_dict(self.obj)
        else:
            # for non-pp objects, e.g. tuple
            try:
                return class_(self.obj, **self.d)
            except ValueError:
                data = json.loads(self.obj)
                df = pd.DataFrame(columns=self.d["columns"])
                for d in data["features"]:
                    idx = int(d["id"])
                    for prop, val in d["properties"].items():
                        df.at[idx, prop] = val
                    # for geom, val in d["geometry"].items():
                    #     df.at[idx, geom] = val
                return df

    if GEOPANDAS_INSTALLED:
        @from_serializable.register(class_name='GeoDataFrame', module_name='geopandas.geodataframe')
        def GeoDataFrame(self):
            fs = json.loads(self.obj)
            # for some reason, the id is not parsed as dataframe id, this is a workaround
            if isinstance(fs, dict) and fs.get("type") == "FeatureCollection":
                features_lst = fs["features"]
            else:
                raise TypeError("not a FeatureCollection")
            for i, feat in enumerate(features_lst):
                fs["features"][i]["properties"]["id"] = feat["id"]

            df = geopandas.GeoDataFrame.from_features(fs, crs=self.d['crs'])
            if "id" in df:
                df.set_index(df['id'].values.astype(numpy.int64), inplace=True)
                df.drop(columns=["id"], inplace=True)
            else:
                df.set_index(df.index.values.astype(numpy.int64), inplace=True)
            df = df.reindex(columns=self.d['columns'])

            # df.astype changes geodataframe to dataframe -> _preserve_dtypes fixes it
            _preserve_dtypes(df, dtypes=self.d["dtype"])
            return df

    if SHAPELY_INSTALLED:
        @from_serializable.register(module_name='shapely')
        def shapely(self):
            return shapely.geometry.shape(self.obj)


class PPJSONDecoder(json.JSONDecoder):
    def __init__(self, **kwargs):
        # net = pandapowerNet.__new__(pandapowerNet)
        #        net = create_empty_network()
        deserialize_pandas = kwargs.pop('deserialize_pandas', True)
        empty_dict_like_object = kwargs.pop('empty_dict_like_object', None)
        registry_class = kwargs.pop("registry_class", FromSerializableRegistry)
        ignore_unknown_objects = kwargs.pop("ignore_unknown_objects", False)
        super_kwargs = {"object_hook": partial(pp_hook,
                                               deserialize_pandas=deserialize_pandas,
                                               empty_dict_like_object=empty_dict_like_object,
                                               registry_class=registry_class,
                                               ignore_unknown_objects=ignore_unknown_objects)}
        super_kwargs.update(kwargs)
        super().__init__(**super_kwargs)


def pp_hook(d, deserialize_pandas=True, empty_dict_like_object=None,
            registry_class=FromSerializableRegistry, ignore_unknown_objects=False):
    try:
        if '_module' in d and '_class' in d:
            if 'pandas' in d['_module'] and not deserialize_pandas:
                return json.dumps(d)
            elif "_object" in d:
                obj = d.pop('_object')
            elif "_state" in d:
                obj = d['_state']
                if '_init' in obj:
                    del obj['_init']
                return obj  # backwards compatibility
            else:
                # obj = {"_init": d, "_state": dict()}  # backwards compatibility
                obj = {key: val for key, val in d.items() if key not in ['_module', '_class']}
            fs = registry_class(obj, d, pp_hook, ignore_unknown_objects)

            fs.class_name = d.pop('_class', '')
            fs.module_name = d.pop('_module', '')
            fs.empty_dict_like_object = empty_dict_like_object
            return fs.from_serializable()
        else:
            return d
    except TypeError:
        logger.debug('Loading your grid raised a TypeError. %s raised this exception' % d)
        return d


def encrypt_string(s, key, compress=True):
    missing_packages = numpy.array(["cryptography", "hashlib", "base64"])[~numpy.array([
        cryptography_INSTALLED, hashlib_INSTALLED, base64_INSTALLED])]
    if len(missing_packages):
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", missing_packages)
    key_base = hashlib.sha256(key.encode())
    key = base64.urlsafe_b64encode(key_base.digest())
    cipher_suite = Fernet(key)

    s = s.encode()
    if compress:
        import zlib
        s = zlib.compress(s)
    s = cipher_suite.encrypt(s)
    s = s.decode()
    return s


def decrypt_string(s, key):
    missing_packages = numpy.array(["cryptography", "hashlib", "base64"])[~numpy.array([
        cryptography_INSTALLED, hashlib_INSTALLED, base64_INSTALLED])]
    if len(missing_packages):
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", missing_packages)
    key_base = hashlib.sha256(key.encode())
    key = base64.urlsafe_b64encode(key_base.digest())
    cipher_suite = Fernet(key)

    s = s.encode()
    s = cipher_suite.decrypt(s)
    if zlib_INSTALLED:
        s = zlib.decompress(s)
    s = s.decode()
    return s


class JSONSerializableClass(object):
    json_excludes = ["self", "__class__"]

    def __init__(self, **kwargs):
        pass

    def to_json(self):
        """
        Each controller should have this method implemented. The resulting json string should be
        readable by the controller's from_json function and by the function add_ctrl_from_json in
        control_handler.
        """
        return json.dumps(self.to_dict(), cls=PPJSONEncoder)

    def to_dict(self):
        def consider_callable(value):
            if callable(value) and value.__class__ in (types.MethodType, types.FunctionType):
                if value.__class__ == types.MethodType and _findclass(value) is not None:
                    return with_signature(value, value.__name__, obj_module=_findclass(value))
                return with_signature(value, value.__name__)
            return value

        d = {key: consider_callable(val) for key, val in self.__dict__.items()
             if key not in self.json_excludes}
        return d

    def add_to_net(self, net, element, index=None, column="object", overwrite=False,
                   preserve_dtypes=False, fill_dict=None):
        if element not in net:
            net[element] = pd.DataFrame(columns=[column])
        if index is None:
            index = get_free_id(net[element])
        if index in net[element].index.values:
            obj = net[element].object.at[index]
            if overwrite or not isinstance(obj, JSONSerializableClass):
                logger.info("Updating %s with index %s" % (element, index))
            else:
                raise UserWarning("%s with index %s already exists" % (element, index))

        dtypes = None
        if preserve_dtypes:
            dtypes = net[element].dtypes

        if fill_dict is not None:
            for k, v in fill_dict.items():
                net[element].at[index, k] = v
        net[element].at[index, column] = self

        if preserve_dtypes:
            _preserve_dtypes(net[element], dtypes)

        return index

    def equals(self, other):
        # todo: can this method be removed?
        warn("JSONSerializableClass: the attribute 'equals' is deprecated "
             "and will be removed in the future. Use the '__eq__' method instead, "
             "by directly comparing the objects 'a == b'. "
             "To check if two variables point to the same object, use 'a is b'", DeprecationWarning)

        logger.warning("JSONSerializableClass: the attribute 'equals' is deprecated "
                       "and will be removed in the future. Use the '__eq__' method instead, "
                       "by directly comparing the objects 'a == b'. "
                       "To check if two variables point to the same object, use 'a is b'")

        class UnequalityFound(Exception):
            pass

        def check_equality(obj1, obj2):
            if isinstance(obj1, (ndarray, generic)) or isinstance(obj2, (ndarray, generic)):
                unequal = True
                if equal(obj1, obj2):
                    unequal = False
                elif anynp(isnan(obj1)):
                    if allclose(obj1, obj2, atol=0, rtol=0, equal_nan=True):
                        unequal = False
                if unequal:
                    raise UnequalityFound
            elif not isinstance(obj2, type(obj1)):
                raise UnequalityFound
            elif isinstance(obj1, pandapowerNet):
                pass
            elif isinstance(obj1, pd.DataFrame):
                if len(obj1) > 0:
                    try:
                        assert_frame_equal(obj1, obj2)
                    except:
                        raise UnequalityFound
            elif isinstance(obj2, pd.Series):
                if len(obj1) > 0:
                    try:
                        assert_series_equal(obj1, obj2)
                    except:
                        raise UnequalityFound
            elif isinstance(obj1, dict):
                check_dictionary_equality(obj1, obj2)
            elif obj1 != obj1 and obj2 != obj2:
                pass
            elif callable(obj1):
                check_callable_equality(obj1, obj2)
            elif obj1 != obj2:
                try:
                    if not (isnan(obj1) and isnan(obj2)):
                        raise UnequalityFound
                except:
                    raise UnequalityFound

        def check_dictionary_equality(obj1, obj2):
            if set(obj1.keys()) != set(obj2.keys()):
                raise UnequalityFound
            for key in obj1.keys():
                if key != "_init":
                    check_equality(obj1[key], obj2[key])

        def check_callable_equality(obj1, obj2):
            if isinstance(obj1, weakref.ref) and isinstance(obj2, weakref.ref):
                return
            if str(obj1) != str(obj2):
                raise UnequalityFound

        if isinstance(other, self.__class__):
            try:
                check_equality(self.__dict__, other.__dict__)
                return True
            except UnequalityFound:
                return False
        else:
            return False

    @classmethod
    def from_dict(cls, d):
        obj = JSONSerializableClass.__new__(cls)
        obj.__dict__.update(d)
        return obj

    @classmethod
    def from_json(cls, json_string):
        d = json.loads(json_string, cls=PPJSONDecoder)
        return cls.from_dict(d)

    def __eq__(self, other):
        """
        comparing class name and attributes instead of class object address directly.
        This allows more flexibility,
        e.g. when the class definition is moved to a different module.
        Checking isinstance(other, self.__class__) for an early return without calling DeepDiff.
        There is still a risk that the implementation details of the methods can differ
        if the classes are from different modules.
        The comparison is based on comparing dictionaries of the classes.
        To this end, the dictionary comparison library deepdiff is used for recursive comparison.
        """
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        else:
            d = DeepDiff(self.__dict__, other.__dict__, ignore_nan_inequality=True,
                         significant_digits=6, math_epsilon=1e-6, ignore_private_variables=False)
            return len(d) == 0

    def __hash__(self):
        # for now we use the address of the object for hash, but we can change it in the future
        # to be based on the attributes e.g. with DeepHash or similar
        return hash(id(self))


def with_signature(obj, val, obj_module=None, obj_class=None):
    if obj_module is None:
        obj_module = obj.__module__.__str__()
    if obj_class is None:
        obj_class = obj.__class__.__name__
    d = {'_module': obj_module, '_class': obj_class, '_object': val}
    if hasattr(obj, 'dtype'):
        d.update({'dtype': str(obj.dtype)})
    return d


@singledispatch
def to_serializable(obj):
    logger.debug('standard case')
    return str(obj)


@to_serializable.register(pandapowerNet)
def json_pandapowernet(obj):
    logger.debug('pandapowerNet')
    net_dict = {k: item for k, item in obj.items() if not k.startswith("_")}
    for k, item in net_dict.items():
        if isinstance(item, str) and '_module' in item:
            net_dict[k] = json.loads(item)
    d = with_signature(obj, net_dict)
    return d


@to_serializable.register(pd.DataFrame)
def json_dataframe(obj):
    logger.debug('DataFrame')
    orient = "split" if not isinstance(obj.index, pd.MultiIndex) and not isinstance(
        obj.columns, pd.MultiIndex) else "columns"
    json_string = obj.to_json(orient=orient, default_handler=to_serializable, double_precision=15)
    d = with_signature(obj, json_string)
    d['orient'] = orient
    if len(obj.columns) > 0 and isinstance(obj.columns[0], str):
        d['dtype'] = obj.dtypes.astype('str').to_dict()

    # store index name (to covert by pandas with orient "split" or "columns" (and totally not with
    # Multiindex))
    if obj.index.name is not None:
        d['index_name'] = obj.index.name
    if obj.columns.name is not None:
        d['column_name'] = obj.columns.name
    if isinstance(obj.index, pd.MultiIndex) and set(obj.index.names) != {None}:
        d['index_names'] = obj.index.names
    if isinstance(obj.columns, pd.MultiIndex) and set(obj.columns.names) != {None}:
        d['column_names'] = obj.columns.names

    # store info that index is of type Multiindex originally
    d['is_multiindex'] = isinstance(obj.index, pd.MultiIndex)
    d['is_multicolumn'] = isinstance(obj.columns, pd.MultiIndex)

    return d


if GEOPANDAS_INSTALLED:
    @to_serializable.register(geopandas.GeoDataFrame)
    def json_geodataframe(obj):
        logger.debug('GeoDataFrame')
        d = with_signature(obj, obj.to_json())
        d.update({'dtype': obj.dtypes.astype('str').to_dict(),
                  'crs': obj.crs, 'columns': obj.columns})
        return d


@to_serializable.register(pd.Series)
def json_series(obj):
    logger.debug('Series')
    orient = "split" if not isinstance(obj.index, pd.MultiIndex) else "index"
    d = with_signature(obj, obj.to_json(orient=orient, default_handler=to_serializable,
                                        double_precision=15))
    d.update({'dtype': str(obj.dtypes), 'orient': orient, 'typ': 'series'})

    # store index name (to covert by pandas with orient "split" or "columns" (and totally not with
    # Multiindex))
    if obj.index.name is not None:
        d['index_name'] = obj.index.name
    if isinstance(obj.index, pd.MultiIndex) and set(obj.index.names) != {None}:
        d['index_names'] = obj.index.names

    # store info that index is of type Multiindex originally
    d['is_multiindex'] = isinstance(obj.index, pd.MultiIndex)

    return d


@to_serializable.register(numpy.ndarray)
def json_array(obj):
    logger.debug("ndarray")
    d = with_signature(obj, list(obj), obj_module='numpy', obj_class='array')
    return d


@to_serializable.register(numpy.integer)
def json_npint(obj):
    logger.debug("integer")
    d = with_signature(obj, int(obj), obj_module="numpy")
    d.pop('dtype')
    return d


@to_serializable.register(numpy.floating)
def json_npfloat(obj):
    logger.debug("floating")
    if numpy.isnan(obj):
        d = with_signature(obj, str(obj), obj_module="numpy")
    else:
        d = with_signature(obj, float(obj), obj_module="numpy")
    d.pop('dtype')
    return d


@to_serializable.register(numpy.bool_)
def json_npbool(obj):
    logger.debug("boolean")
    d = with_signature(obj, "true" if obj else "false", obj_module="numpy")
    d.pop('dtype')
    return d


@to_serializable.register(numbers.Number)
def json_num(obj):
    logger.debug("numbers.Number")
    return str(obj)


@to_serializable.register(complex)
def json_complex(obj):
    logger.debug("complex")
    d = with_signature(obj, str(obj), obj_module='builtins', obj_class='complex')
    d.pop('dtype')
    return d


@to_serializable.register(pd.Index)
def json_pdindex(obj):
    logger.debug("pd.Index")
    return with_signature(obj, list(obj), obj_module='pandas')


@to_serializable.register(bool)
def json_bool(obj):
    logger.debug("bool")
    return "true" if obj else "false"


@to_serializable.register(tuple)
def json_tuple(obj):
    logger.debug("tuple")
    d = with_signature(obj, list(obj), obj_module='builtins', obj_class='tuple')
    return d


@to_serializable.register(set)
def json_set(obj):
    logger.debug("set")
    d = with_signature(obj, list(obj), obj_module='builtins', obj_class='set')
    return d


@to_serializable.register(frozenset)
def json_frozenset(obj):
    logger.debug("frozenset")
    d = with_signature(obj, list(obj), obj_module='builtins', obj_class='frozenset')
    return d


@to_serializable.register(networkx.Graph)
def json_networkx(obj):
    logger.debug("nx graph")
    json_string = json_graph.adjacency_data(obj, attrs={'id': 'json_id', 'key': 'json_key'})
    d = with_signature(obj, json_string, obj_module="networkx")
    return d


@to_serializable.register(JSONSerializableClass)
def controller_to_serializable(obj):
    logger.debug('JSONSerializableClass')
    d = with_signature(obj, obj.to_json())
    return d


def mkdirs_if_not_existent(dir_to_create):
    already_exist = os.path.isdir(dir_to_create)
    os.makedirs(dir_to_create, exist_ok=True)
    return not already_exist


if SHAPELY_INSTALLED:
    @to_serializable.register(shapely.geometry.LineString)
    def json_linestring(obj):
        logger.debug("shapely linestring")
        json_string = shapely.geometry.mapping(obj)
        d = with_signature(obj, json_string, obj_module="shapely")
        return d


    @to_serializable.register(shapely.geometry.Point)
    def json_point(obj):
        logger.debug("shapely Point")
        json_string = shapely.geometry.mapping(obj)
        d = with_signature(obj, json_string, obj_module="shapely")
        return d


    @to_serializable.register(shapely.geometry.Polygon)
    def json_polygon(obj):
        logger.debug("shapely Polygon")
        json_string = shapely.geometry.mapping(obj)
        d = with_signature(obj, json_string, obj_module="shapely")
        return d



import os
import uuid
from copy import deepcopy
from pathlib import Path

import numpy as np
import pandas as pd

from pandapower.auxiliary import _init_runpp_options, _add_dcline_gens, LoadflowNotConverged
from pandapower.create import create_ext_grid, create_bus, create_impedance, create_transformer_from_parameters, \
    create_load
from pandapower.diagnostic import diagnostic
from pandapower.file_io import to_json
from pandapower.pd2ppc import _pd2ppc
from pandapower.pf.ppci_variables import _get_pf_variables_from_ppci
from pandapower.pf.run_newton_raphson_pf import _get_numba_functions, _get_Y_bus
from pandapower.run import _passed_runpp_parameters, runpp
from pandapower.toolbox.element_selection import pp_elements, get_connected_elements_dict, branch_element_bus_dict
from pandapower.toolbox.grid_modification import drop_lines, drop_trafos, drop_measurements_at_elements, \
    drop_controllers_at_buses

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)
home = str(Path.home())
desktop = os.path.join(home, "Desktop")
impedance_columns = ["from_bus", "to_bus", "rft_pu", "xft_pu", "rtf_pu", "xtf_pu", "gf_pu", "bf_pu", "gt_pu", "bt_pu"]


def _runpp_except_voltage_angles(net, **kwargs):
    if "calculate_voltage_angles" not in kwargs or not kwargs["calculate_voltage_angles"]:
        runpp(net, **kwargs)
    else:
        try:
            runpp(net, **kwargs)
        except LoadflowNotConverged:
            kwargs1 = deepcopy(kwargs)
            kwargs1["calculate_voltage_angles"] = False
            runpp(net, **kwargs1)
            logger.warning("In grid equivalent generation, the power flow did converge only without"
                           " calculate_voltage_angles.")
    return net


def add_ext_grids_to_boundaries(net, boundary_buses, adapt_va_degree=False,
                                runpp_fct=_runpp_except_voltage_angles,
                                calc_volt_angles=True, allow_net_change_for_convergence=False,
                                **kwargs):
    """
    adds ext_grids for the given network. If the bus results are
    available, ext_grids are created according to the given bus results;
    otherwise, ext_grids are created with vm_pu=1 and va_degreee=0
    """
    orig_slack_gens = net.gen.index[net.gen.slack]
    buses_to_add_ext_grids = set(boundary_buses) - set(net.ext_grid.bus[net.ext_grid.in_service]) \
                             - set(net.gen.bus[net.gen.in_service & net.gen.slack])
    res_buses = set(
        net.res_bus.index[~net.res_bus[["vm_pu", "va_degree"]].isnull().any(axis=1)])
    btaegwr = list(buses_to_add_ext_grids & res_buses)
    add_eg = []
    vms = pd.Series(np.ones(len(buses_to_add_ext_grids)),
                    index=buses_to_add_ext_grids)
    vas = pd.Series(np.zeros(len(buses_to_add_ext_grids)),
                    index=buses_to_add_ext_grids)
    vms.loc[btaegwr] = net.res_bus.vm_pu.loc[btaegwr]
    vms.loc[pd.Index(net.gen.bus.loc[net.gen.in_service]).intersection(vms.index)] = \
        net.gen.vm_pu.loc[net.gen.in_service & net.gen.bus.isin(vms.index) &
                          ~net.gen.bus.duplicated()].values  # avoid
    # different vm_pu setpoints at same buses
    vas.loc[btaegwr] = net.res_bus.va_degree.loc[btaegwr]

    for ext_bus, vm, va in zip(buses_to_add_ext_grids, vms, vas):
        add_eg += [create_ext_grid(net, ext_bus,
                                   vm, va, name="assist_ext_grid")]
        new_bus = create_bus(net, net.bus.vn_kv[ext_bus], name="assist_bus")
        create_impedance(net, ext_bus, new_bus, 1e6, 1e6, net.sn_mva,
                         name="assist_impedance")

    # works fine if there is only one slack in net:
    if adapt_va_degree and net.gen.slack.any() and net.ext_grid.shape[0]:
        slack_buses = net.gen.bus.loc[net.gen.slack]
        net.gen.slack = False
        try:
            runpp_fct(net, calculate_voltage_angles=calc_volt_angles,
                      max_iteration=100, **kwargs)
        except LoadflowNotConverged as e:
            if allow_net_change_for_convergence:

                # --- various fix trials

                # --- trail 1 -> massive change of data (switch sign of impedances)
                imp_neg = net.impedance.index[(net.impedance.xft_pu < 0)]
                imp_neg = net.impedance[["xft_pu"]].loc[imp_neg].sort_values("xft_pu").index
                for no, idx in enumerate(imp_neg):
                    net.impedance.loc[idx, ["rft_pu", "rtf_pu", "xft_pu", "xtf_pu"]] *= -1
                    try:
                        runpp_fct(net, calculate_voltage_angles=True, max_iteration=100, **kwargs)
                        logger.warning("The sign of these impedances were changed to enable a power"
                                       f" flow: {imp_neg[:no]}")
                        break
                    except LoadflowNotConverged as e:
                        pass

                if not net.converged:
                    net.impedance.loc[imp_neg, ["rft_pu", "rtf_pu", "xft_pu", "xtf_pu"]] *= -1

                    # --- trail 2 -> increase impedance values to avoid close to zero values
                    changes = False
                    for col in ["xtf_pu", "xft_pu"]:
                        is2small = net.impedance[col].abs() < 5e-6
                        changes |= is2small.any()
                        sign = np.sign(net.impedance[col].values[is2small])
                        net.impedance[col].loc[is2small] = sign * 5e-6
                    if changes:
                        try:
                            runpp_fct(net, calculate_voltage_angles=calc_volt_angles,
                                      max_iteration=100, **kwargs)
                            logger.warning("Reactances of these impedances has been increased to "
                                           f"enable a power flow: {is2small}")
                        except LoadflowNotConverged as e:
                            diag = diagnostic(net)
                            print(net)
                            print(diag.keys())
                            to_json(net, os.path.join(desktop, "diverged_net.json"))
                            raise LoadflowNotConverged(e)
                    else:
                        diag = diagnostic(net)
                        print(net)
                        print(diag.keys())
                        to_json(net, os.path.join(desktop, "diverged_net.json"))
                        raise LoadflowNotConverged(e)
            else:
                raise LoadflowNotConverged(e)

        va = net.res_bus.va_degree.loc[slack_buses]
        va_ave = va.sum() / va.shape[0]
        net.ext_grid.va_degree.loc[add_eg] -= va_ave
        runpp_fct(net, calculate_voltage_angles=calc_volt_angles,
                  max_iteration=100, **kwargs)
    return orig_slack_gens


def drop_internal_branch_elements(net, internal_buses, branch_elements=None):
    """
    This function drops all branch elements which have 'internal_buses' connected at all sides of
    the branch element (e.g. for lines at 'from_bus' and 'to_bus').
    """
    bebd = branch_element_bus_dict()
    if branch_elements is not None:
        bebd = {elm: bus_types for elm,
        bus_types in bebd.items() if elm in branch_elements}
    for elm, bus_types in bebd.items():
        n_elms = net[elm].shape[0]
        if n_elms:
            should_be_dropped = np.ones((n_elms,), dtype=bool)
            for bus_type in bus_types:
                should_be_dropped &= net[elm][bus_type].isin(internal_buses)
            idx_to_drop = net[elm].index[should_be_dropped]
            if elm == "line":
                drop_lines(net, idx_to_drop)
            elif "trafo" in elm:
                drop_trafos(net, idx_to_drop, table=elm)
            else:
                net[elm] = net[elm].drop(idx_to_drop)


def calc_zpbn_parameters(net, boundary_buses, all_external_buses, slack_as="gen",
                         existing_shift_degree=False):
    """
    The function calculats the parameters for zero power balance network

    OUTPUT:
        **Z** - impedance between essential buses and new buses

        **v** - voltage at new buses
    """
    #    runpp_fct(net, calculate_voltage_angles=True)
    be_buses = boundary_buses + all_external_buses
    if ((net.trafo.hv_bus.isin(be_buses)) & (net.trafo.shift_degree != 0)).any() \
            or ((net.trafo3w.hv_bus.isin(be_buses)) & \
                ((net.trafo3w.shift_mv_degree != 0) | (net.trafo3w.shift_lv_degree != 0))).any():
        existing_shift_degree = True
        logger.info("Transformers with non-zero shift-degree are existed," +
                    " they could cause small inaccuracy.")
    # creata dataframe to collect the current injections of the external area
    nb_ext_buses = len(all_external_buses)
    S = pd.DataFrame(np.zeros((nb_ext_buses, 15)), dtype=complex)
    S.columns = ["ext_bus", "v_m", "v_cpx", "gen_integrated", "gen_separate",
                 "load_integrated", "load_separate", "sgen_integrated",
                 "sgen_separate", "sn_load_separate", "sn_load_integrated",
                 "sn_sgen_separate", "sn_sgen_integrated", "sn_gen_separate",
                 "sn_gen_integrated"]

    k, ind = 0, 0
    if slack_as == "gen":
        elements = set([("load", "res_load", "load_separate", "sn_load_separate", -1),
                        ("sgen", "res_sgen", "sgen_separate", "sn_sgen_separate", 1),
                        ("gen", "res_gen", "gen_separate", "sn_gen_separate", 1),
                        ("ext_grid", "res_ext_grid", "gen_separate", "sn_gen_separate", 1)])

    elif slack_as == "load":
        elements = set([("load", "res_load", "load_separate", "sn_load_separate", -1),
                        ("sgen", "res_sgen", "sgen_separate", "sn_sgen_separate", 1),
                        ("gen", "res_gen", "gen_separate", "sn_gen_separate", 1),
                        ("ext_grid", "res_ext_grid", "load_separate", "sn_load_separate", 1)])

    for i in all_external_buses:
        for ele, res_ele, power, sn, sign in elements:
            if i in net[ele].bus.values and net[ele].in_service[net[ele].bus == i].values.any():
                ind = list(net[ele].index[net[ele].bus == i].values)
                # act. values --> ref. values:
                S.loc[k, power] += sum(net[res_ele].p_mw[ind].values * sign) / net.sn_mva + \
                                   1j * sum(net[res_ele].q_mvar[ind].values *
                                            sign) / net.sn_mva
                S.loc[k, sn] = sum(net[ele].sn_mva[ind].values) + \
                               1j * 0 if ele != "ext_grid" else 1e6 + 1j * 0
                S[power.replace('_separate', '_integrated')] += S[power][k]
                S[sn.replace('_separate', '_integrated')] += S[sn][k]
        S.loc[k, 'ext_bus'] = all_external_buses[k]
        S.loc[k, 'v_m'] = net.res_bus.vm_pu[i]
        S.loc[k, 'v_cpx'] = S.v_m[k] * \
                            np.exp(1j * net.res_bus.va_degree[i] * np.pi / 180)
        k = k + 1

    # create dataframe to calculate the impedance of the ZPBN-network
    Y = pd.DataFrame(np.zeros((nb_ext_buses, 10)), dtype=complex)
    Y.columns = ["ext_bus", "load_ground", "load_integrated_total", "load_separate_total",
                 "gen_ground", "gen_integrated_total", "gen_separate_total",
                 "sgen_ground", "sgen_integrated_total", "sgen_separate_total"]
    v = pd.DataFrame(dtype=complex)
    Y.ext_bus, v["ext_bus"] = all_external_buses, all_external_buses

    for elm in ["load", "gen", "sgen"]:
        if existing_shift_degree:
            Y[elm + "_ground"] = (S[elm + "_separate"].values / S.v_cpx.values).conjugate() / \
                                 S.v_cpx.values
        else:
            Y[elm + "_ground"] = S[elm + "_separate"].values.conjugate() / \
                                 np.square(S.v_m)
        I_elm_integrated_total = sum((S[elm + "_separate"].values /
                                      S.v_cpx.values).conjugate())
        if I_elm_integrated_total == 0:
            Y[elm + "_integrated_total"] = float("nan")
        else:
            vm_elm_integrated_total = S[elm + "_integrated"][0] / \
                                      I_elm_integrated_total.conjugate()
            if existing_shift_degree:
                Y[elm + "_integrated_total"] = (-S[elm + "_integrated"][0] / \
                                                vm_elm_integrated_total).conjugate() / \
                                               vm_elm_integrated_total
            else:
                Y[elm + "_integrated_total"] = -S[elm + "_integrated"][0].conjugate() / \
                                               np.square(abs(vm_elm_integrated_total))
        Y[elm + "_separate_total"] = -Y[elm + "_ground"]
        if elm == "gen" and any(S.gen_separate):
            v["gen_integrated_vm_total"] = abs(vm_elm_integrated_total)
            v["gen_separate_vm_total"] = S.v_m

    Z = -1 / Y
    Z.ext_bus = all_external_buses

    # determine original external bus limits
    limits = pd.DataFrame([], index=range(nb_ext_buses), columns=[
        "min_vm_pu", "max_vm_pu", "ext_bus"])
    for limit in ["min_vm_pu", "max_vm_pu"]:
        if limit in net.bus.columns:
            limits[limit] = net.bus[limit].loc[all_external_buses].values
    limits["ext_bus"] = all_external_buses

    return Z, S, v, limits


def _ensure_unique_boundary_bus_names(net, boundary_buses):
    """ This function ad a unique name to each bounary bus. The original
        boundary bus names are retained.
    """
    assert "name_equivalent" not in net.bus.columns.tolist()
    net.bus["name_equivalent"] = "uuid"
    net.bus.loc[boundary_buses, "name_equivalent"] = ["Boundary bus " + str(uuid.uuid1()) for _ in
                                                      boundary_buses]


def drop_assist_elms_by_creating_ext_net(net, elms=None):
    """
    This function drops the assist elements by creating external nets.
    """
    if elms is None:
        elms = ["ext_grid", "bus", "impedance"]
    for elm in elms:
        target_elm_idx = net[elm].index[net[elm].name.astype(str).str.contains(
            "assist_" + elm, na=False, regex=False)]
        net[elm] = net[elm].drop(target_elm_idx)
        if net["res_" + elm].shape[0]:
            res_target_elm_idx = net["res_" +
                                     elm].index.intersection(target_elm_idx)
            net["res_" + elm] = net["res_" + elm].drop(res_target_elm_idx)

    if "name_equivalent" in net.bus.columns.tolist():
        net.bus = net.bus.drop(columns=["name_equivalent"])


def build_ppc_and_Ybus(net):
    """ This function build ppc and gets the Ybus of given network without
    runing power flow calculation
    """
    loc = locals()
    loc['kwargs'] = {}
    _init_runpp_options(net,
                        algorithm='nr',
                        calculate_voltage_angles="auto",
                        init="auto",
                        max_iteration="auto",
                        tolerance_mva=1e-8,
                        trafo_model="t",
                        trafo_loading="current",
                        enforce_q_lims=False,
                        check_connectivity=True,
                        voltage_depend_loads=True,
                        consider_line_temperature=False,
                        passed_parameters=_passed_runpp_parameters(loc))

    ppc, ppci = _pd2ppc(net)
    net["_ppc"] = ppc
    makeYbus, pfsoln = _get_numba_functions(ppci, net["_options"])
    baseMVA, bus, gen, branch, *_, V0, ref_gens = _get_pf_variables_from_ppci(ppci)
    _, Ybus, _, _ = _get_Y_bus(ppci, net["_options"], makeYbus, baseMVA, bus, branch)

    net._ppc["internal"]["Ybus"] = Ybus


def drop_measurements_and_controllers(net, buses, skip_controller=False):
    """This function drops the measurements of the given buses.
    Also, the related controller parameters will be removed. """
    # --- dropping measurements
    if len(net.measurement):
        drop_measurements_at_elements(net, "bus", idx=buses)
        lines = net.line.index[(net.line.from_bus.isin(buses)) &
                               (net.line.from_bus.isin(buses))]
        drop_measurements_at_elements(net, "line", idx=lines)
        trafos = net.trafo3w.index[(net.trafo3w.hv_bus.isin(buses)) &
                                   (net.trafo3w.mv_bus.isin(buses)) &
                                   (net.trafo3w.lv_bus.isin(buses))]
        drop_measurements_at_elements(net, "trafo", idx=trafos)
        trafo3ws = net.trafo3w.index[(net.trafo3w.hv_bus.isin(buses)) &
                                     (net.trafo3w.mv_bus.isin(buses)) &
                                     (net.trafo3w.lv_bus.isin(buses))]
        drop_measurements_at_elements(net, "trafo3w", idx=trafo3ws)

    # --- dropping controller
    drop_controllers_at_buses(net, buses)


def ensure_origin_id(net, elms=None):
    """
    Ensures completely filled column 'origin_id' in every pp element.
    """
    if elms is None:
        elms = pp_elements()

    for elm in elms:
        if "origin_id" not in net[elm].columns:
            net[elm]["origin_id"] = pd.Series([None] * net[elm].shape[0], dtype=object)
        idxs = net[elm].index[net[elm].origin_id.isnull()]
        net[elm].loc[idxs, "origin_id"] = ["%s_%i_%s" % (elm, idx, str(uuid.uuid4())) for idx in idxs]


def drop_and_edit_cost_functions(net, buses, drop_cost, add_origin_id,
                                 check_unique_elms_name=True):
    """
    This function drops the ploy_cost/pwl_cost data
    related to the given buses.
    """
    for cost_elm in ["poly_cost", "pwl_cost"]:
        if len(net[cost_elm]):
            cost_backup = net[cost_elm].copy()
            # drop poly_cost and pwl_cost
            if drop_cost:
                net[cost_elm]["bus"] = None
                for elm in set(net[cost_elm].et.values):
                    idx = net[cost_elm].element.index[(net[cost_elm].et == elm) &
                                                      (net[cost_elm].element.isin(net[elm].index))]
                    net[cost_elm].loc[idx, "bus"] = net[elm].bus.loc[net[cost_elm].element.loc[
                        idx]].values
                to_drop = net[cost_elm].index[net[cost_elm].bus.isin(buses) |
                                              net[cost_elm].bus.isnull()]
                net[cost_elm] = net[cost_elm].drop(to_drop)

            # add origin_id to cost df and corresponding elms
            if add_origin_id:
                ensure_origin_id(net, elms=set(net[cost_elm].et.values))
                if "et_origin_id" not in net[cost_elm].columns:
                    net[cost_elm]["et_origin_id"] = None
                    net[cost_elm]["origin_idx"] = None
                    net[cost_elm]["origin_seq"] = None
                for elm in set(net[cost_elm].et.values):
                    idx = net[cost_elm].index[net[cost_elm].et == elm]
                    net[cost_elm].loc[idx, "et_origin_id"] = net[elm].origin_id.loc[net[
                        cost_elm].element.loc[idx]].values
                    net[cost_elm].loc[idx, "origin_idx"] = idx
                    net[cost_elm].loc[idx, "origin_seq"] = [cost_backup.index.tolist().index(t) for t in idx]


def match_cost_functions_and_eq_net(net, boundary_buses, eq_type):
    """
    This function makes the element indices in poly_cost/pwl_cost and the
    new element indecies after merging to match.
    """
    for cost_elm in ["poly_cost", "pwl_cost"]:
        if len(net[cost_elm]):
            if "ward" not in eq_type:
                net[cost_elm] = net[cost_elm].sort_values(by=["origin_seq"])
                net[cost_elm].index = net[cost_elm]["origin_idx"].values
                for pc in net[cost_elm].itertuples():
                    new_idx = net[pc.et].index[
                        net[pc.et].origin_id == pc.et_origin_id].values
                    net[cost_elm].loc[pc.Index, 'element'] = new_idx[0]
            net[cost_elm] = net[cost_elm].drop(columns=["bus", "et_origin_id", "origin_idx", "origin_seq"])


def _check_network(net):
    """
    This function will perfoms some checks and modifications on the given grid model.
    """
    # --- check invative elements
    if net.res_bus.vm_pu.isnull().any():
        logger.info("There are some inactive buses. It is suggested to remove "
                    "them using 'pandapower.drop_inactive_elements()' "
                    "before starting the grid equivalent calculation.")

    # --- check dclines
    if "dcline" in net and len(net.dcline.query("in_service")) > 0:
        _add_dcline_gens(net)
        dcline_index = net.dcline.index.values
        net.dcline.loc[dcline_index, 'in_service'] = False
        logger.info(f"replaced dcline {dcline_index} by gen elements")

    # --- check controller names
    if len(net.controller):
        for i in net.controller.index:
            et = net.controller.object[i].__dict__.get("element")
            if et is not None and len(net[et]) != len(set(net[et].name.values)):
                raise ValueError("if controllers are used, please give a name for every "
                                 "element (" + et + "), and make sure the name is unique.")


def get_boundary_vp(net_eq, bus_lookups):
    v_boundary = net_eq.ext_grid
    p_boundary = net_eq.res_ext_grid.values + net_eq.res_bus[["p_mw", "q_mvar"]].loc[
        bus_lookups["bus_lookup_pd"]["b_area_buses"]].values
    p_boundary = pd.DataFrame(p_boundary, index=bus_lookups["bus_lookup_pd"]["b_area_buses"],
                              columns=["p_mw", "q_mvar"])
    return v_boundary, p_boundary


def adaptation_phase_shifter(net, v_boundary, p_boundary):
    target_buses = list(v_boundary.bus.values)
    phase_errors = v_boundary.va_degree.values - \
                   net.res_bus.va_degree[target_buses].values
    vm_errors = v_boundary.vm_pu.values - \
                net.res_bus.vm_pu[target_buses].values
    # p_errors = p_boundary.p_mw.values - \
    #     net.res_bus.p_mw[target_buses].values
    # q_errors = p_boundary.q_mvar.values - \
    #     net.res_bus.q_mvar[target_buses].values
    # print(q_errors)
    for idx, lb in enumerate(target_buses):
        if abs(vm_errors[idx]) > 1e-6 and abs(vm_errors[idx]) > 1e-6:
            hb = create_bus(net, net.bus.vn_kv[lb] * (1 - vm_errors[idx]),
                            name="phase_shifter_adapter_" + str(lb))
            elm_dict = get_connected_elements_dict(net, lb)
            for e, e_list in elm_dict.items():
                for i in e_list:
                    name = str(net[e].name[i])
                    if "eq_" not in name and "_integrated_" not in name and \
                            "_separate_" not in name:
                        if e in ["impedance", "line"]:
                            if net[e].from_bus[i] == lb:
                                net[e].from_bus[i] = hb
                            else:
                                net[e].to_bus[i] = hb
                        elif e == "trafo":
                            if net[e].hv_bus[i] == lb:
                                net[e].loc[i, 'hv_bus'] = hb
                            else:
                                net[e].lv_bus[i] = hb
                        elif e == "trafo3w":
                            if net[e].hv_bus[i] == lb:
                                net[e].hv_bus[i] == hb
                            elif net[e].mv_bus[i] == lb:
                                net[e].mv_bus[i] == hb
                            else:
                                net[e].lv_bus[i] == lb
                        elif e in ["bus", "load", "sgen", "gen", "shunt", "ward", "xward"]:
                            pass
                        else:
                            net[e].loc[i, 'bus'] = hb
            create_transformer_from_parameters(net, hb, lb, 1e5,
                                               net.bus.vn_kv[hb] * (1 - vm_errors[idx]),
                                               net.bus.vn_kv[lb],
                                               vkr_percent=0, vk_percent=100,
                                               pfe_kw=.0, i0_percent=.0,
                                               # shift_degree=-phase_errors[idx],
                                               tap_step_degree=-phase_errors[idx],
                                               # tap_phase_shifter=True,
                                               name="phase_shifter_adapter_" + str(lb))
        # pp.create_load(net, lb, -p_errors[idx], -q_errors[idx],
        #                name="phase_shifter_adapter_"+str(lb))
    # runpp_fct(net, calculate_voltage_angles=True)
    return net


def replace_motor_by_load(net, all_external_buses):
    """
    replace the 'external' motors by loads. The name is modified.
    e.g., "equivalent_MotorName_3" ("equivalent"+"original name"+"original index")
    """
    motors = net.motor.index[net.motor.bus.isin(all_external_buses)]
    for mi, m in net.motor.loc[motors].iterrows():
        p_mech = m.pn_mech_mw / (m.efficiency_percent / 100)
        p_mw = p_mech * m.loading_percent / 100 * m.scaling
        s = p_mw / m.cos_phi
        q_mvar = np.sqrt(s ** 2 - p_mw ** 2)
        li = create_load(net, m.bus, p_mw, q_mvar, sn_mva=s, scalling=m.scaling,
                         in_service=m.in_service, name="equivalent_" + str(m["name"]) + "_" + str(mi))
        p = p_mw if not np.isnan(net.res_bus.vm_pu[m.bus]) and m.in_service else 0.0
        q = q_mvar if not np.isnan(net.res_bus.vm_pu[m.bus]) and m.in_service else 0.0
        net.res_load.loc[li] = p, q
    net.motor = net.motor.drop(motors)
    net.res_motor = net.res_motor.drop(motors)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


try:
    from matplotlib.colors import ListedColormap, BoundaryNorm, LinearSegmentedColormap, Normalize, LogNorm
    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False

import numpy as np


def cmap_discrete(cmap_list):
    """
    Can be used to create a discrete colormap.

    INPUT:
        - cmap_list (list) - list of tuples, where each tuple represents one range. Each tuple has
                             the form of ((from, to), color).

    OUTPUT:
        - cmap - matplotlib colormap

        - norm - matplotlib norm object

    EXAMPLE:
        >>> from pandapower.plotting import cmap_discrete, create_line_collection, draw_collections
        >>> from pandapower.networks import mv_oberrhein
        >>> net = mv_oberrhein("generation")
        >>> cmap_list = [((0, 10), "green"), ((10, 30), "yellow"), ((30, 100), "red")]
        >>> cmap, norm = cmap_discrete(cmap_list)
        >>> lc = create_line_collection(net, cmap=cmap, norm=norm)
        >>> draw_collections([lc])
    """
    if not MATPLOTLIB_INSTALLED:
        raise UserWarning("install matplotlib to use this function")
    cmap_colors = []
    boundaries = []
    last_upper = None
    for (lower, upper), color in cmap_list:
        if last_upper is not None and lower != last_upper:
            raise ValueError("Ranges for colormap must be continuous")
        cmap_colors.append(color)
        boundaries.append(lower)
        last_upper = upper
    boundaries.append(upper)
    cmap = ListedColormap(cmap_colors)
    norm = BoundaryNorm(boundaries, cmap.N)
    return cmap, norm


def cmap_continuous(cmap_list):
    """
    Can be used to create a continuous colormap.

    INPUT:
        - cmap_list (list) - list of tuples, where each tuple represents one color. Each tuple has
                             the form of (center, color). The colorbar is a linear segmentation of
                             the colors between the centers.

    OUTPUT:
        - cmap - matplotlib colormap

        - norm - matplotlib norm object

    EXAMPLE:
        >>> from pandapower.plotting import cmap_continuous, create_bus_collection, draw_collections
        >>> from pandapower.networks import mv_oberrhein
        >>> net = mv_oberrhein("generation")
        >>> cmap_list = [(0.97, "blue"), (1.0, "green"), (1.03, "red")]
        >>> cmap, norm = cmap_continuous(cmap_list)
        >>> bc = create_bus_collection(net, size=70, cmap=cmap, norm=norm)
        >>> draw_collections([bc])
    """
    if not MATPLOTLIB_INSTALLED:
        raise UserWarning("install matplotlib to use this function")
    min_loading = cmap_list[0][0]
    max_loading = cmap_list[-1][0]
    cmap_colors = [((loading-min_loading)/(max_loading - min_loading), color) for
                 (loading, color) in cmap_list]
    cmap = LinearSegmentedColormap.from_list('name', cmap_colors)
    norm = Normalize(min_loading, max_loading)
    return cmap, norm


def cmap_logarithmic(min_value, max_value, colors):
    """
        Can be used to create a logarithmic colormap. The colormap itself has a linear segmentation of
        the given colors. The values however will be matched to the colors based on a logarithmic
        normalization (c.f. matplotlib.colors.LogNorm for more information on how the logarithmic
        normalization works).

        \nPlease note: {There are numerous ways of how a logarithmic scale might
                        be created, the intermediate values on the scale are created automatically based on the minimum
                        and maximum given values in analogy to the LogNorm. Also, the logarithmic colormap can only be
                        used with at least 3 colors and increasing values which all have to be above 0.}

        INPUT:
            **min_value** (float) - the minimum value of the colorbar

            **max_value** (float) - the maximum value for the colorbar

            **colors** (list) - list of colors to be used for the colormap

        OUTPUT:
            **cmap** - matplotlib colormap

            **norm** - matplotlib norm object

        EXAMPLE:

        >>> from pandapower.plotting import cmap_logarithmic, create_bus_collection, draw_collections
        >>> from pandapower.networks import mv_oberrhein
        >>> net = mv_oberrhein("generation")
        >>> min_value, max_value = 1.0, 1.03
        >>> colors = ["blue", "green", "red"]
        >>> cmap, norm = cmap_logarithmic(min_value, max_value, colors)
        >>> bc = create_bus_collection(net, size=70, cmap=cmap, norm=norm)
        >>> draw_collections([bc])

    """

    num_values = len(colors)
    if num_values < 2:
        raise UserWarning("Cannot create a logarithmic colormap less than 2 colors.")
    if min_value <= 0:
        raise UserWarning("The minimum value must be above 0.")
    if max_value <= min_value:
        raise UserWarning("The upper bound must be larger than the lower bound.")
    values = np.arange(num_values + 1)
    diff = (max_value - min_value) / (num_values - 1)
    values = (np.log(min_value + values * diff) - np.log(min_value)) \
             / (np.log(max_value) - np.log(min_value))
    cmap = LinearSegmentedColormap.from_list("name", list(zip(values, colors)))
    norm = LogNorm(min_value, max_value)
    return cmap, norm



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import os

from typing_extensions import deprecated

from pandapower.plotting.geo import convert_crs

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)


def _on_map_test(x, y):
    """
    checks if bus_geodata can be located on a map using geopy
    """
    try:
        from geopy.geocoders import Nominatim
        from geopy.exc import GeocoderTimedOut
        geolocator = Nominatim(user_agent="pandapower_user_mapboxplot")

    except ImportError:
        # if geopy is not available there will be no geo-coordinates check
        # therefore if geo-coordinates are not real and user sets on_map=True, an empty map will be plot!
        raise ImportError(
            'Geo-coordinates check cannot be performed because geopy package not available \n\t--> '
            'if geo-coordinates are not in lat/lon format an empty plot may appear...'
        )
    try:
        location = geolocator.reverse(f"{x}, {y}", language='en-US')
    except GeocoderTimedOut:
        logger.error("Existing net geodata cannot be geo-located: possible reason: geo-data not in lat/long ->"
                     "try geo_data_to_latlong(net, projection) to transform geodata to lat/long!")
    else:
        if location.address is None:
            return False
    return True


@deprecated('geo_data_to_latlong is deprecated and will be removed shortly, use pandapower.geo.convert_crs instead')
def geo_data_to_latlong(net, projection):
    """
    Transforms network's geodata (in `net.bus_geodata` and `net.line_geodata`) from specified projection to lat/long (WGS84).

    INPUT:
        **net** (pandapowerNet) - The pandapower network

        **projection** (String) - projection from which geodata are transformed to lat/long. some examples

                - "epsg:31467" - 3-degree Gauss-Kruger zone 3
                - "epsg:2032" - NAD27(CGQ77) / UTM zone 18N
                - "epsg:2190" - Azores Oriental 1940 / UTM zone 26N
    """
    convert_crs(net, epsg_in=projection.split(':')[1], epsg_out=4326)


def set_mapbox_token(token):
    from pandapower.__init__ import pp_dir
    path = os.path.join(pp_dir, "plotting", "plotly")
    filename = os.path.join(path, 'mapbox_token.txt')
    with open(filename, "w") as mapbox_file:
        mapbox_file.write(token)


def _get_mapbox_token():
    from pandapower.__init__ import pp_dir
    path = os.path.join(pp_dir, "plotting", "plotly")
    filename = os.path.join(path, 'mapbox_token.txt')
    with open(filename, "r") as mapbox_file:
        return mapbox_file.read()



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import copy

from pandapower.auxiliary import get_free_id
from pandapower.control.util.auxiliary import \
    drop_same_type_existing_controllers, log_same_type_existing_controllers
from pandapower.io_utils import JSONSerializableClass

try:
    import pandaplan.core.pplog as pplog
except:
    import logging as pplog

logger = pplog.getLogger(__name__)


class BasicCtrl(JSONSerializableClass):
    """
    Base-Class of all controllable elements within a network.
    """

    def __init__(self, container, index=None, **kwargs):
        super().__init__()
        # add oneself to net, creating the ['controller'] DataFrame, if necessary
        if index is None:
            index = get_free_id(container.controller)
        self.index = index

    def __repr__(self):
        rep = "This " + self.__class__.__name__ + " has the following parameters: \n"

        for member in ["index", "json_excludes"]:
            rep += ("\n" + member + ": ").ljust(20)
            d = locals()
            exec('value = self.' + member, d)
            rep += str(d['value'])

        return rep

    def __str__(self):
        s = self.__class__.__name__
        return s

    def __getstate__(self):
        state = copy.deepcopy(self.__dict__)

        for attr in self.json_excludes:
            try:
                del state[attr]
            except:
                continue

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def time_step(self, container, time):
        """
        It is the first call in each time step, thus suited for things like
        reading profiles or prepare the controller for the next control step.

        .. note:: This method is ONLY being called during time-series simulation!
        """
        pass

    def initialize_control(self, container):
        """
        Some controller require extended initialization in respect to the
        current state of the net (or their view of it). This method is being
        called after an initial loadflow but BEFORE any control strategies are
        being applied.

        This method may be interesting if you are aiming for a global
        controller or if it has to be aware of its initial state.
        """
        pass

    def is_converged(self, container):
        """
        This method calculated whether or not the controller converged. This is
        where any target values are being calculated and compared to the actual
        measurements. Returns convergence of the controller.
        """
        logger.warning("Method is_converged() has not been overwritten "
                       "(and will always return True)!")
        return True

    def control_step(self, container):
        """
        If the is_converged method returns false, the control_step will be
        called. In other words: if the controller did not converge yet, this
        method should implement actions that promote convergence e.g. adapting
        actuating variables and writing them back to the data structure.
        """
        pass

    def repair_control(self, container):
        """
        Some controllers can cause net to not converge. In this case, they can implement a method to
        try and catch the load flow error by altering some values in net, for example load scaling.
        This method is being called in the except block in run_control.
        Either implement this in a controller that is likely to cause the error, or define
        a special "load flow police" controller for your use case
        """
        pass

    def restore_init_state(self, container):
        """
        Some controllers manipulate values in net and then restore them back to initial values, e.g.
        DistributedSlack.
        This method should be used for such a purpose because it is executed in the except block of
        run_control to make sure that the net condition is restored even if load flow calculation
        doesn't converge
        """
        pass

    def finalize_control(self, container):
        """
        Some controller require extended finalization. This method is being
        called at the end of a loadflow.
        It is a separate method from restore_init_state because it is possible that control
        finalization does not only restore the init state but also something in addition to that,
        that would require the results in net
        """
        pass

    def finalize_step(self, container, time):
        """
        .. note:: This method is ONLY being called during time-series simulation!

        After each time step, this method is being called to clean things up or
        similar. The OutputWriter is a class specifically designed to store
        results of the loadflow. If the ControlHandler.output_writer got an
        instance of this class, it will be called before the finalize step.
        """
        pass

    def set_active(self, container, in_service):
        """
        Sets the controller in or out of service
        """
        container.controller.loc[self.index, 'in_service'] = in_service


    def level_reset(self, prosumer):
        pass


class Controller(BasicCtrl):
    """
    Base-Class of all controllable elements within a network.
    """

    def __init__(self, net, in_service=True, order=0, level=0, index=None, recycle=False,
                 drop_same_existing_ctrl=False, initial_run=True, overwrite=False,
                 matching_params=None, **kwargs):
        super(Controller, self).__init__(net, index, **kwargs)
        self.matching_params = dict() if matching_params is None else matching_params
        # add oneself to net, creating the ['controller'] DataFrame, if necessary
        # even though this code is repeated in JSONSerializableClass, it is necessary because of how drop_same_existing_controller works
        # it is still needed in JSONSerializableClass because it is used for characteristics
        if index is None and "controller" in net.keys():
            index = get_free_id(net.controller)
        self.index = self.add_controller_to_net(net=net, in_service=in_service, initial_run=initial_run,
                                                order=order, level=level, index=index, recycle=recycle,
                                                drop_same_existing_ctrl=drop_same_existing_ctrl,
                                                overwrite=overwrite, matching_params=matching_params, **kwargs)

    def add_controller_to_net(self, net, in_service, initial_run, order, level, index, recycle,
                              drop_same_existing_ctrl, overwrite, **kwargs):
        """
        adds the controller to net['controller'] dataframe.

        INPUT:
            **in_service** (bool) - in service status

            **order** (int) - order

            **index** (int) - index

            **recycle** (bool) - if controller needs a new bbm (ppc, Ybus...) or if it can be used \
                                 with prestored values. This is mostly needed for time series \
                                 calculations

        """
        if drop_same_existing_ctrl:
            drop_same_type_existing_controllers(net, type(self), index=index, **kwargs)
        else:
            log_same_type_existing_controllers(net, type(self), index=index, **kwargs)

        # use base class method to raise an error if the object is in DF and overwrite = False
        # if the index is None, the base class is in charge of obtaining the next free index in the data frame
        fill_dict = {"in_service": in_service, "initial_run": initial_run, "recycle": recycle,
                     "order": order, "level": level}
        added_index = super().add_to_net(net=net, element='controller', index=index, overwrite=overwrite,
                           fill_dict=fill_dict, preserve_dtypes=True)
        return added_index

    def time_step(self, net, time):
        super().time_step(net, time)

    def initialize_control(self, net):
        super().initialize_control(net)

    def is_converged(self, net):
        return super().is_converged(net)

    def control_step(self, net):
        super().control_step(net)

    def repair_control(self, net):
        super().repair_control(net)

    def restore_init_state(self, net):
        super().restore_init_state(net)

    def finalize_control(self, net):
        super().finalize_control(net)

    def finalize_step(self, net, time):
        super().finalize_step(net, time)

    def set_active(self, net, in_service):
        super().set_active(net, in_service)

    def set_recycle(self, net):
        """
        Checks the recyclability of this controller and changes the recyclability of the control handler if
        necessary. With this a faster time series calculation can be achieved since not everything must be
        recalculated.

        Beware: Setting recycle wrong can mess up your results. Set it to False in init if in doubt!
        """
        # checks what can be reused from this controller - default is False in base controller
        pass

